"""
Instrument driver for the Edwards Turbo Instrument Controller.
"""
from enum import Enum
import logging
import re
from typing import Dict, NamedTuple, Tuple

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)

# command and response strings
GENERAL_CMD_FORMAT = "!C{obj_id} {msg}"
GENERAL_RESP_FORMAT = "*C{obj_id} "
SETUP_CMD_FORMAT = "!S{obj_id} {data}"
SETUP_RESP_FORMAT = "*S{obj_id} "
QUERY_SETUP_CMD_FORMAT = "?S{obj_id}"
QUERY_SETUP_RESP_ERR_FORMAT = "*S{obj_id} "
QUERY_SETUP_RESP_FORMAT = "=S{obj_id} "
QUERY_VALUE_CMD_FORMAT = "?V{obj_id}"
QUERY_VALUE_RESP_ERR_FORMAT = "*V{obj_id} "
QUERY_VALUE_RESP_FORMAT = "=V{obj_id} "


class EdwardsVacuum_TIC_ErrorCode(Enum):
    """
    Error code for the instrument.
    """
    NO_ERROR = 0
    INVALID_COMMAND_FOR_OBJECT_ID = 1
    INVALID_QUERY_OR_COMMAND = 2
    MISSING_PARAMETER = 3
    PARAMETER_OUT_OF_RANGE = 4
    INVALID_COMMAND_IN_STARTUP_STAGE = 5
    DATA_CHECKSUM_ERROR = 6
    EEPROM_READ_OR_WRITE_ERROR = 7
    OPERATION_TOOK_TOO_LONG = 8
    INVALID_CONFIG_ID = 9


class EdwardsVacuum_TIC_Priority(Enum):
    """
    Priority level for instrument.
    """
    OK = 0
    WARNING = 1
    ALARM = 2
    ALARM_2 = 3


class EdwardsVacuum_TIC_AlertId(Enum):
    """
    Alert ID for instrument.
    """
    NO_ALERT = 0
    ADC_FAULT = 1
    ADC_NOT_READ = 2
    OVER_RANGE = 3
    UNDER_RANGE = 4
    ADC_INVALID = 5
    NO_GAUGE = 6
    UNKNOWN = 7
    NOT_SUPPORTED = 8
    NEW_ID = 9
    OVER_RANGE_2 = 10
    UNDER_RANGE_2 = 11
    OVER_RANGE_3 = 12
    ION_EM_TIMEOUT = 13
    NOT_STRUCK = 14
    FILAMENT_FAIL = 15
    MAG_FAIL = 16
    STRIKER_FAIL = 17
    NOT_STUCK = 18
    FILAMENT_FAIL_2 = 19
    CAL_ERROR = 20
    INITIALISING = 21
    EMISSION_ERROR = 22
    OVER_PRESSURE = 23
    ASG_CANT_ZERO = 24
    RAMPUP_TIMEOUT = 25
    DROOP_TIMEOUT = 26
    RUN_HOURS_HIGH = 27
    SC_INTERLOCK = 28
    ID_VOLTS_ERROR = 29
    SERIAL_ID_FAIL = 30
    UPLOAD_ACTIVE = 31
    DX_FAULT = 32
    TEMP_ALERT = 33
    SYSI_INHIBIT = 34
    EXT_INHIBIT = 35
    TEMP_INHIBIT = 36
    NO_READING = 37
    NO_MESSAGE = 38
    NOV_FAILURE = 39
    UPLOAD_TIMEOUT = 40
    DOWNLOAD_FAILED = 41
    NO_TUBE = 42
    USE_GAUGES_4_TO_6 = 43
    DEGAS_INHIBITED = 44
    IGC_INHIBITED = 45
    BROWNOUT_OR_SHORT = 46
    SERVICE_DUE = 47


class EdwardsVacuum_TIC_State(Enum):
    """
    State of a device in the instrument.
    """
    OFF = 0
    OFF_GOING_ON = 1
    ON_GOING_OFF_SHUTDOWN = 2
    ON_GOING_OFF_NORMAL = 3
    ON = 4


class EdwardsVacuum_TIC_PumpState(Enum):
    """
    Pump state of instrument.
    """
    STOPPED = 0
    STARTING_DELAY = 1
    STOPPING_SHORT_DELAY = 2
    STOPPING_NORMAL_DELAY = 3
    RUNNING = 4
    ACCELERATING = 5
    FAULT_BRAKING = 6
    BRAKING = 7


class EdwardsVacuum_TIC_GaugeState(Enum):
    """
    Gauge state of the instrument.
    """
    GAUGE_NOT_CONNECTED = 0
    GAUGE_CONNECTED = 1
    NEW_GAUGE_ID = 2
    GAUGE_CHANGED = 3
    GAUGE_IN_ALERT = 4
    OFF = 5
    STRIKING = 6
    INITIALISING = 7
    CALIBRATING = 8
    ZEROING = 9
    DEGASSING = 10
    ON = 11
    INHIBITED = 12


# unit types
UNIT_TYPES = {
    66: "V",
    59: "Pa",
    81: "%"
}

# object IDs
TIC_STATUS = 902
TURBO_PUMP = 904
TURBO_SPEED = 905
TURBO_POWER = 906
BACKING_PUMP = 910
BACKING_SPEED = 911
BACKING_POWER = 912
SYSTEM = 933

PRESSURE_GAUGE_MAP = {
    1: 913,
    2: 914,
    3: 915
}

RELAY_MAP = {
    1: 916,
    2: 917,
    3: 918
}


class EdwardsVacuum_TIC_PumpStateResponse(NamedTuple):
    """
    State of the pump.

    Attributes:
        state:      State of the pump.
        alert:      Alert given by the system.
        priority:   Priority level of the system.
    """
    state: EdwardsVacuum_TIC_PumpState
    alert: EdwardsVacuum_TIC_AlertId
    priority: EdwardsVacuum_TIC_Priority


class EdwardsVacuum_TIC_PumpSpeedResponse(NamedTuple):
    """
    Speed of the pump.

    Attributes:
        speed:      Speed as a percentage of the maximum value [0-100%].
        alert:      Alert given by the system.
        priority:   Priority level of the system.
    """
    speed: float
    alert: EdwardsVacuum_TIC_AlertId
    priority: EdwardsVacuum_TIC_Priority


class EdwardsVacuum_TIC_PumpPowerResponse(NamedTuple):
    """
    Speed of the pump.

    Attributes:
        power:      Power of the pump in Watts.
        alert:      Alert given by the system.
        priority:   Priority level of the system.
    """
    power: float
    alert: EdwardsVacuum_TIC_AlertId
    priority: EdwardsVacuum_TIC_Priority


class EdwardsVacuum_TIC_GaugePressureResponse(NamedTuple):
    """
    Pressure read by the gauge.

    Attributes:
        pressure:   Pressure.
        unit:       Unit of the value.
        state:      State of the gauge.
        alert:      Alert given by the system.
        priority:   Priority level of the system.
    """
    pressure: float
    unit: str
    state: EdwardsVacuum_TIC_GaugeState
    alert: EdwardsVacuum_TIC_AlertId
    priority: EdwardsVacuum_TIC_Priority


class EdwardsVacuum_TIC_StateResponse(NamedTuple):
    """
    State of a device of the instrument.

    Attributes:
        state:      State of the device.
        alert:      Alert given by the system.
        priority:   Priority level of the system.
    """
    state: EdwardsVacuum_TIC_State
    alert: EdwardsVacuum_TIC_AlertId
    priority: EdwardsVacuum_TIC_Priority


class EdwardsVacuum_TIC_StatusResponse(NamedTuple):
    """
    Status of the instument.

    Attributes:
        turbo_pump_state:   State of the turbo pump.
        backing_pump_state: State of the backing pump.
        gauge_1_state:      State of gauge 1.
        gauge_2_state:      State of gauge 2.
        gauge_3_state:      State of gauge 3.
        relay_1_state:      State of relay 1.
        relay_2_state:      State of relay 2.
        relay_3_state:      State of relay 3.
        alert:              Alert given by the system.
        priority:           Priority level of the system.
    """
    turbo_pump_state: EdwardsVacuum_TIC_PumpState
    backing_pump_state: EdwardsVacuum_TIC_PumpState
    gauge_1_state: EdwardsVacuum_TIC_GaugeState
    gauge_2_state: EdwardsVacuum_TIC_GaugeState
    gauge_3_state: EdwardsVacuum_TIC_GaugeState
    relay_1_state: EdwardsVacuum_TIC_State
    relay_2_state: EdwardsVacuum_TIC_State
    relay_3_state: EdwardsVacuum_TIC_State
    alert: EdwardsVacuum_TIC_AlertId
    priority: EdwardsVacuum_TIC_Priority


class EdwardsVacuum_TIC_SystemOnOffSetupConfigResponse(NamedTuple):
    """
    System on/off setup config of the instrument.

    Attributes:
        turbo_pump_setup:   On/off setup config of the turbo pump. First entry of tuple is the on config and
                            second is the off config.
        backing_pump_setup: On/off setup config of the backing pump. First entry of tuple is the on config and
                            second is the off config.
        gauge_1_setup:      On/off setup config of gauge 1. First entry of tuple is the on config and
                            second is the off config.
        gauge_2_setup:      On/off setup config of gauge 2. First entry of tuple is the on config and
                            second is the off config.
        gauge_3_setup:      On/off setup config of gauge 3. First entry of tuple is the on config and
                            second is the off config.
        relay_1_setup:      On/off setup config of relay 1. First entry of tuple is the on config and
                            second is the off config.
        relay_2_setup:      On/off setup config of relay 2. First entry of tuple is the on config and
                            second is the off config.
        relay_3_setup:      On/off setup config of relay 3. First entry of tuple is the on config and
                            second is the off config.
    """
    turbo_pump_setup: Tuple[bool, bool]
    backing_pump_setup: Tuple[bool, bool]
    gauge_1_setup: Tuple[bool, bool]
    gauge_2_setup: Tuple[bool, bool]
    gauge_3_setup: Tuple[bool, bool]
    relay_1_setup: Tuple[bool, bool]
    relay_2_setup: Tuple[bool, bool]
    relay_3_setup: Tuple[bool, bool]


class Edwards_TurboInstrumentController(QMI_Instrument):
    """Instrument driver for the Edwards Turbo Instrument Controller."""

    # default response timeout in seconds
    DEFAULT_RESPONSE_TIMEOUT = 5.0

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str
                 ) -> None:
        """Initialise driver.

        Parameters:
            transport:  QMI transport descriptor to connect to the instrument.
        """
        super().__init__(context, name)
        self._timeout = self.DEFAULT_RESPONSE_TIMEOUT
        self._transport_str = transport
        self._transport = create_transport(transport)
        self._scpi_protocol = ScpiProtocol(self._transport,
                                           command_terminator="\r",
                                           response_terminator="\r",
                                           default_timeout=self._timeout)

    def _parse_response(self, raw_response: str) -> str:
        """
        Parse the received response by doing a pattern check and extracting the message string

        Parameters:
            raw_response: The raw response received from the instrument.

        Returns:
            Message sent by the instrument.
        """
        resp, response_data = raw_response.split(" ", 1)
        # match the first 2 characters of the string to find what type of response it is
        err_resp_pattern = "[*][SCV]"
        response_pattern = "[=][SCV]"
        if re.match(err_resp_pattern, raw_response) is not None:
            err_code = int(response_data)
            if err_code != 0:
                # when the response is *C with a code of 1, that is not an error, but a response
                if not (err_code == 1 and resp[1] == "C"):
                    raise QMI_InstrumentException(f"Error {EdwardsVacuum_TIC_ErrorCode(err_code)}")
            return response_data
        elif re.match(response_pattern, raw_response) is None:
            raise QMI_InstrumentException("The response format did not match any valid response formats")
        return response_data

    def _get_pump_state(self, obj_id: int) -> EdwardsVacuum_TIC_PumpStateResponse:
        """
        Get the state of the pump.

        Parameters:
            obj_id: The object id to send the message to.

        Returns:
            An instance of EdwardsVacuum_TIC_PumpStateResponse.
        """
        _logger.info("Getting pump status for instrument [%s]", self._name)
        cmd = QUERY_VALUE_CMD_FORMAT.format(obj_id=obj_id)
        resp = [int(x) for x in self._parse_response(self._scpi_protocol.ask(cmd)).split(";")]
        return EdwardsVacuum_TIC_PumpStateResponse(
            state=EdwardsVacuum_TIC_PumpState(resp[0]),
            alert=EdwardsVacuum_TIC_AlertId(resp[1]),
            priority=EdwardsVacuum_TIC_Priority(resp[2])
        )

    def _get_pump_speed(self, obj_id: int) -> EdwardsVacuum_TIC_PumpSpeedResponse:
        """
        Get the speed of the pump as a percentage of the maximum speed.

        Parameters:
            obj_id: The object id to send the message to.

        Returns:
            An instance of EdwardsVacuum_TIC_PumpSpeedResponse.
        """
        _logger.info("Getting the pump speed for instrument [%s]", self._name)
        cmd = QUERY_VALUE_CMD_FORMAT.format(obj_id=obj_id)
        resp = self._parse_response(self._scpi_protocol.ask(cmd)).split(";")
        return EdwardsVacuum_TIC_PumpSpeedResponse(
            speed=float(resp[0]),
            alert=EdwardsVacuum_TIC_AlertId(int(resp[1])),
            priority=EdwardsVacuum_TIC_Priority(int(resp[2]))
        )

    def _turn_on_off(self, obj_id: int, on_off: bool) -> None:
        """
        Turn on or off an object.

        Parameters:
            obj_id: The object id to send the message to.
            on_off: False means off and True means on.
        """
        cmd = GENERAL_CMD_FORMAT.format(obj_id=obj_id, msg=int(on_off))
        self._parse_response(self._scpi_protocol.ask(cmd))

    def _get_pump_power(self, obj_id: int) -> EdwardsVacuum_TIC_PumpPowerResponse:
        """
        Get the power of the pump.

        Parameters:
            obj_id: The object to send the message to.

        Returns:
            An instance of EdwardsVacuum_TIC_PumpPowerResponse.
        """
        _logger.info("Getting pump power for instrument [%s]", self._name)
        cmd = QUERY_VALUE_CMD_FORMAT.format(obj_id=obj_id)
        resp = self._parse_response(self._scpi_protocol.ask(cmd)).split(";")
        return EdwardsVacuum_TIC_PumpPowerResponse(
            power=float(resp[0]),
            alert=EdwardsVacuum_TIC_AlertId(int(resp[1])),
            priority=EdwardsVacuum_TIC_Priority(int(resp[2]))
        )

    def _get_pressure(self, obj_id: int) -> EdwardsVacuum_TIC_GaugePressureResponse:
        """
        Get the pressure from a gauge.

        Parameters:
            obj_id: The object to send the message to.

        Returns:
            An instance of EdwardsVacuum_TIC_GaugePressureResponse.
        """
        cmd = QUERY_VALUE_CMD_FORMAT.format(obj_id=obj_id)
        resp = self._parse_response(self._scpi_protocol.ask(cmd)).split(";")
        return EdwardsVacuum_TIC_GaugePressureResponse(
            pressure=float(resp[0]),
            unit=UNIT_TYPES[int(resp[1])],
            state=EdwardsVacuum_TIC_GaugeState(int(resp[2])),
            alert=EdwardsVacuum_TIC_AlertId(int(resp[3])),
            priority=EdwardsVacuum_TIC_Priority(int(resp[4]))
        )

    def _get_state(self, obj_id: int) -> EdwardsVacuum_TIC_StateResponse:
        """
        Get the state of the provided object id.

        Parameters:
            obj_id: The object id.

        Returns:
            An instance of EdwardsVacuum_TIC_StateResponse.
        """
        cmd = QUERY_VALUE_CMD_FORMAT.format(obj_id=obj_id)
        resp = self._parse_response(self._scpi_protocol.ask(cmd)).split(";")
        return EdwardsVacuum_TIC_StateResponse(
            state=EdwardsVacuum_TIC_State(int(resp[0])),
            alert=EdwardsVacuum_TIC_AlertId(int(resp[1])),
            priority=EdwardsVacuum_TIC_Priority(int(resp[2]))
        )

    @rpc_method
    def open(self) -> None:
        _logger.info("Opening connection instrument [%s]", self._name)
        self._check_is_closed()
        self._transport.open()
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("Closing connection to instrument [%s]", self._name)
        self._check_is_open()
        self._transport.close()
        super().close()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """
        Read instrument type and version and return QMI_InstrumentIdentification instance.

        Returns:
            Instrument identification.
        """
        _logger.info("Getting info for instrument [%s]", self._name)
        cmd = QUERY_SETUP_CMD_FORMAT.format(obj_id=TIC_STATUS)
        instr_info = self._parse_response(self._scpi_protocol.ask(cmd)).split(";")
        return QMI_InstrumentIdentification(vendor="Edwards",
                                            model=instr_info[0],
                                            version=instr_info[1],
                                            serial=instr_info[2])

    @rpc_method
    def get_tic_status(self) -> EdwardsVacuum_TIC_StatusResponse:
        """
        Get the status of the turbo pump, backing pump and pressure gauge
        attached to the TIC.

        Returns:
            An instance of EdwardsVacuum_TIC_StatusResponse.
        """
        _logger.info("Getting status for instrument [%s]", self._name)
        cmd = QUERY_VALUE_CMD_FORMAT.format(obj_id=TIC_STATUS)
        resp = [int(x) for x in self._parse_response(self._scpi_protocol.ask(cmd)).split(";")]
        return EdwardsVacuum_TIC_StatusResponse(
            turbo_pump_state=EdwardsVacuum_TIC_PumpState(resp[0]),
            backing_pump_state=EdwardsVacuum_TIC_PumpState(resp[1]),
            gauge_1_state=EdwardsVacuum_TIC_GaugeState(resp[2]),
            gauge_2_state=EdwardsVacuum_TIC_GaugeState(resp[3]),
            gauge_3_state=EdwardsVacuum_TIC_GaugeState(resp[4]),
            relay_1_state=EdwardsVacuum_TIC_State(resp[5]),
            relay_2_state=EdwardsVacuum_TIC_State(resp[6]),
            relay_3_state=EdwardsVacuum_TIC_State(resp[7]),
            alert=EdwardsVacuum_TIC_AlertId(resp[8]),
            priority=EdwardsVacuum_TIC_Priority(resp[9])
        )

    @rpc_method
    def get_turbo_pump_state(self) -> EdwardsVacuum_TIC_PumpStateResponse:
        """
        Get the state of the turbo pump.

        Returns:
            An instance of EdwardsVacuum_TIC_PumpStateResponse.
        """
        return self._get_pump_state(TURBO_PUMP)

    @rpc_method
    def get_backing_pump_state(self) -> EdwardsVacuum_TIC_PumpStateResponse:
        """
        Get the state of the backing pump.

        Returns:
            An instance of EdwardsVacuum_TIC_PumpStateResponse.
        """
        return self._get_pump_state(BACKING_PUMP)

    @rpc_method
    def turn_on_turbo_pump(self) -> None:
        """
        Turn on the turbo pump.
        """
        _logger.info("Turning on the turbo pump for instrument [%s]", self._name)
        self._turn_on_off(TURBO_PUMP, True)

    @rpc_method
    def turn_off_turbo_pump(self) -> None:
        """
        Turn off the turbo pump.
        """
        _logger.info("Turning off the turbo pump for instrument [%s]", self._name)
        self._turn_on_off(TURBO_PUMP, False)

    @rpc_method
    def turn_on_backing_pump(self) -> None:
        """
        Turn on the turbo pump.
        """
        _logger.info("Turning on the backing pump for instrument [%s]", self._name)
        self._turn_on_off(BACKING_PUMP, True)

    @rpc_method
    def turn_off_backing_pump(self) -> None:
        """
        Turn off the turbo pump.
        """
        _logger.info("Turning off the backing pump for instrument [%s]", self._name)
        self._turn_on_off(BACKING_PUMP, False)

    @rpc_method
    def get_turbo_pump_speed(self) -> EdwardsVacuum_TIC_PumpSpeedResponse:
        """
        Get the speed of the turbo pump as a percentage of the maximum speed.

        Returns:
            An instance of EdwardsVacuum_TIC_PumpSpeedResponse.
        """
        return self._get_pump_speed(TURBO_SPEED)

    @rpc_method
    def get_backing_pump_speed(self) -> EdwardsVacuum_TIC_PumpSpeedResponse:
        """
        Get the speed of the backing pump as a percentage of the maximum speed.

        Returns:
            An instance of EdwardsVacuum_TIC_PumpStateResponse.
        """
        return self._get_pump_speed(BACKING_SPEED)

    @rpc_method
    def get_turbo_pump_power(self) -> EdwardsVacuum_TIC_PumpPowerResponse:
        """
        Get the power of the turbo pump.

        Returns:
            An instance of EdwardsVacuum_TIC_PumpPowerResponse.
        """
        return self._get_pump_power(TURBO_POWER)

    @rpc_method
    def get_backing_pump_power(self) -> EdwardsVacuum_TIC_PumpPowerResponse:
        """
        Get the power of the backing pump.

        Returns:
            An instance of EdwardsVacuum_TIC_PumpPowerResponse.
        """
        return self._get_pump_power(BACKING_POWER)

    @rpc_method
    def get_pressure(self, gauge_num: int) -> EdwardsVacuum_TIC_GaugePressureResponse:
        """
        Get the pressure from the pressure gauge.

        Parameters:
            gauge_num:  Gauge to address.

        Returns:
            An instance of EdwardsVacuum_TIC_GaugePressureResponse.
        """
        _logger.info("Getting pressure from gauge %s for instrument [%s]", gauge_num, self._name)
        return self._get_pressure(PRESSURE_GAUGE_MAP[gauge_num])

    @rpc_method
    def get_relay_state(self, relay_num: int) -> EdwardsVacuum_TIC_StateResponse:
        """
        Get the state of the relay.

        Parameters:
            relay_num:  Relay to address.

        Returns:
            An instance of EdwardsVacuum_TIC_StateResponse.
        """
        _logger.info("Getting state of relay %s for instrument [%s]", relay_num, self._name)
        return self._get_state(RELAY_MAP[relay_num])

    @rpc_method
    def turn_on_relay(self, relay_num: int) -> None:
        """
        Turn on the relay.

        Parameters:
            relay_num:  Relay to address.
        """
        _logger.info("Turning on relay %s for instrument [%s]", relay_num, self._name)
        self._turn_on_off(RELAY_MAP[relay_num], True)

    @rpc_method
    def turn_off_relay(self, relay_num: int) -> None:
        """
        Turn off the relay.
        """
        _logger.info("Turning off relay %s for instrument [%s]", relay_num, self._name)
        self._turn_on_off(RELAY_MAP[relay_num], False)

    @rpc_method
    def get_system_on_off_setup_config(self) -> EdwardsVacuum_TIC_SystemOnOffSetupConfigResponse:
        """
        Get the on/off setup config of the system.

        Returns:
            An instance of EdwardsVacuum_TIC_SystemOnOffSetupConfigResponse.
        """
        _logger.info("Getting on/off setup config of system for instrument [%s]", self._name)
        cmd = QUERY_SETUP_CMD_FORMAT.format(obj_id=SYSTEM)
        resp = self._parse_response(self._scpi_protocol.ask(cmd)).split(";")
        bool_resp = [bool(int(i)) for i in resp]
        return EdwardsVacuum_TIC_SystemOnOffSetupConfigResponse(
            turbo_pump_setup=(bool_resp[1], bool_resp[2]),
            backing_pump_setup=(bool_resp[4], bool_resp[5]),
            gauge_1_setup=(bool_resp[7], bool_resp[8]),
            gauge_2_setup=(bool_resp[10], bool_resp[11]),
            gauge_3_setup=(bool_resp[13], bool_resp[14]),
            relay_1_setup=(bool_resp[16], bool_resp[17]),
            relay_2_setup=(bool_resp[19], bool_resp[20]),
            relay_3_setup=(bool_resp[22], bool_resp[23]))

    @rpc_method
    def set_system_on_off_setup_config(self, config: Dict[int, Tuple[bool, bool]]) -> None:
        """
        Set the on/off setup config of the system.

        Parameters:
            config: The object to setup. The first entry in the tuple is the object id.
                    the second is a tuple for the on off setup.
        """
        _logger.info("Setting on/off setup config of system for instrument [%s]", self._name)
        data = ""
        for obj_id, setup in config.items():
            data += f"{obj_id};{int(setup[0])};{int(setup[1])};"
        cmd = SETUP_CMD_FORMAT.format(obj_id=SYSTEM, data=data)
        # remove the last ";"" from the string and send that as the command
        self._scpi_protocol.ask(cmd[:-1])

    @rpc_method
    def get_system_state(self) -> EdwardsVacuum_TIC_StateResponse:
        """
        Get the state of the system.

        Returns:
            An instance of EdwardsVacuum_TIC_StateResponse.
        """
        _logger.info("Getting state of system for instrument [%s]", self._name)
        return self._get_state(SYSTEM)

    @rpc_method
    def turn_on_system(self) -> None:
        """
        Turn on the system.
        """
        _logger.info("Turning on system for instrument [%s]", self._name)
        self._turn_on_off(SYSTEM, True)

    @rpc_method
    def turn_off_system(self) -> None:
        """
        Turn off the system.
        """
        _logger.info("Turning off system for instrument [%s]", self._name)
        self._turn_on_off(SYSTEM, False)



"""Implementation of the QMI_Context class.
"""

import logging
import os
import pathlib
import random
import re
import selectors
import socket
import string
import threading
import time
import warnings
import atexit
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple, Type

import qmi

from qmi.core.config_defs import CfgQmi, CfgContext
from qmi.core.exceptions import QMI_UsageException, QMI_DuplicateNameException, QMI_UnknownNameException, \
                                QMI_ConfigurationException, QMI_InvalidOperationException, QMI_WrongThreadException
from qmi.core.messaging import MessageRouter, QMI_Message, QMI_MessageHandlerAddress, QMI_MessageHandler
from qmi.core.rpc import QMI_RpcObject, QMI_RpcProxy, RpcObjectManager, rpc_method, RpcObjectDescriptor, \
                         make_interface_descriptor, QMI_LockTokenDescriptor
from qmi.core.pubsub import SignalManager, QMI_SignalReceiver
from qmi.core.instrument import QMI_Instrument
from qmi.core.task import QMI_Task, QMI_TaskRunner
from qmi.core.udp_responder_packets import unpack_qmi_udp_packet, QMI_UdpResponderContextInfoRequestPacket, \
                                           QMI_UdpResponderContextInfoResponsePacket
from qmi.core.util import is_valid_object_name, format_address_and_port, AtomicCounter


# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


# Global variable counting the number of active contexts.
_active_context_counter = AtomicCounter()


@atexit.register
def _check_active_contexts() -> None:
    """Check that there are no active QMI contexts when the program exits.

    This function runs just before the Python program exits.
    It emits a warning if there are still active QMI contexts.

    Programs that use QMI should always stop their QMI context before
    the program ends, to ensure that resources such as network connections
    and background threads are properly cleaned up.
    """

    count = _active_context_counter.value()
    if count:
        # Note: ResourceWarning is by default not displayed in Python.
        warnings.warn("Still {} active QMI contexts at program exit".format(count), ResourceWarning)


class _UdpPingResponse(NamedTuple):
    """Ping response descriptor."""
    received_timestamp: float
    incoming_address: Tuple[str, int]
    response_packet: QMI_UdpResponderContextInfoResponsePacket


def ping_qmi_contexts(workgroup_name_filter: str,
                      context_name_filter: str = "*",
                      timeout: float = 0.1) -> List[_UdpPingResponse]:
    """Broadcast an info request message to discover all QMI contexts on the network.

    Parameters:
        workgroup_name_filter:  filter on workgroup name.
        context_name_filter:    filter on context name (default: "*").
        timeout:                time to wait for answers (default: 0.1).
    """
    responses = []

    # Prepare a socket.
    udp_socket = socket.socket(family=socket.AF_INET, type=socket.SOCK_DGRAM)
    try:
        # Configure the socket.
        udp_socket.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        address_in = ("", 0)  # the empty string represents INADDR_ANY; port 0 means: give me any port
        udp_socket.bind(address_in)

        # Prepare the outgoing packet.
        request_pkt_id = random.randint(1, 2**64 - 1)
        request_packet = QMI_UdpResponderContextInfoRequestPacket.create(
            request_pkt_id,
            time.time(),
            workgroup_name_filter.encode(),
            context_name_filter.encode()
        )
        address_out = ("<broadcast>", QMI_Context.DEFAULT_UDP_RESPONDER_PORT)
        udp_socket.sendto(request_packet, address_out)  # type: ignore

        # Collect answers.
        wait_until = time.monotonic() + timeout
        with selectors.DefaultSelector() as selector:
            key_udp_packet_received = selector.register(udp_socket, selectors.EVENT_READ)

            while True:
                t_now = time.monotonic()
                if t_now >= wait_until:
                    break

                for key, _ in selector.select(wait_until - t_now):
                    if key is key_udp_packet_received:

                        # A packet is received, determine if it is a response packet/
                        response_packet, incoming_address = udp_socket.recvfrom(4096)
                        response_received_timestamp = time.time()

                        try:
                            unpacked_response_packet = unpack_qmi_udp_packet(response_packet)
                        except BaseException:
                            continue  # ignore unpack failure; probably not a QMI packet

                        if not isinstance(unpacked_response_packet, QMI_UdpResponderContextInfoResponsePacket):
                            continue  # not a response

                        if unpacked_response_packet.request_pkt_id != request_pkt_id:
                            continue  # not a response to our request

                        # The packet is a response to our broadcast.
                        response = _UdpPingResponse(
                            received_timestamp=response_received_timestamp,
                            incoming_address=incoming_address,
                            response_packet=unpacked_response_packet)
                        responses.append(response)

    finally:
        udp_socket.close()

    return responses


class _ContextRpcObject(QMI_RpcObject):
    """Internal class, used to provide information about the local context via RPC."""

    @classmethod
    def get_category(cls) -> Optional[str]:
        return "context"

    @rpc_method
    def get_version(self) -> str:
        """Return the QMI version used by this context."""
        return self._context.get_version()

    @rpc_method
    def get_rpc_object_descriptors(self) -> List[RpcObjectDescriptor]:
        """Get a list RPC descriptors of the objects in this context."""
        return self._context.get_rpc_object_descriptors()

    @rpc_method
    def get_rpc_object_descriptor(self, rpc_object_name: str) -> Optional[RpcObjectDescriptor]:
        """Get an RPC descriptor for the object with the given name."""
        return self._context.get_rpc_object_descriptor(rpc_object_name)

    @rpc_method
    def get_pid(self) -> int:
        """Return process ID of Python program."""
        return os.getpid()

    @rpc_method
    def shutdown_context(self, hard: bool) -> None:
        """Tell the context to shut down.

        A normal (soft) shutdown requires active cooperation from the main
        program. It sets a flag to indicate that a shutdown request was received.
        The main program should actively monitor this flag to perform cleanup
        and exit as soon as the flag is set.

        A hard shutdown will almost always succeed but may leave instruments
        and data files in an unclean state.
        """

        if hard:
            # Exit the process now (this also exits all threads).
            _logger.info("Received hard shutdown request")
            print("Answering external hard-kill request, exiting with exitcode 1.")
            os._exit(1)

        else:
            # Set the shutdown request flag.
            _logger.info("Received soft shutdown request")
            self._context._context_shutdown_requested.set()


class QMI_Context:
    """Represents an application or script within the QMI system.

    The `QMI_Context` is the main entry point to QMI functionality.
    It keeps track of *instruments*, *tasks*, *signals*, *RPC objects*
    and *connections* to other contexts. It provides functionality to create
    and access these objects, and to interact with objects in other contexts.

    Internally, the QMI context will create Python threads to perform certain
    tasks in the background.

    Unless stated otherwise, methods of the `QMI_Context` instance may only
    be called from the thread that created the context. A few specific methods
    may safely be called from any thread; this is explicitly stated in the
    documentation of the method when it applies.

    In most Python programs that use the QMI framework, there is exactly
    one instance of QMI_Context. This instance is created when the application
    calls ``qmi.start(...)`` and destroyed when the application calls ``qmi.stop()``.

    Attributes:
        name: Name of the context (read-only).
        workgroup_name: Name of the QMI workgroup (read-only).
    """

    DEFAULT_UDP_RESPONDER_PORT = 35999

    def __init__(self, name: str, config: Optional[CfgQmi] = None) -> None:
        """Create a new `QMI_Context` instance.

        This is normally done automatically by ``qmi.start(...)``.
        Most applications should not explicitly create a `QMI_Context` instance.

        Parameters:
            name: Name of the context.
                This must be a short string without spaces or strange characters.
                If other contexts will connect to this context, its name must
                be unique within the workgroup (i.e. among all Python processes
                that work together via QMI).
            config: QMI configuration data.
        """

        _logger.debug("Initializing context %r", name)

        # The "active" variable indicates if the context is currently active.
        self._active = False

        # The "used" variable prevents restarting the context after it was previously stopped.
        self._used = False

        # Event which is set to true when the context receives a shutdown request via RPC.
        self._context_shutdown_requested = threading.Event()

        # Program start time (may be used to create log file names etc.).
        self._start_time = time.time()

        if not is_valid_object_name(name):
            raise QMI_UsageException("Invalid context name {!r}".format(name))

        self.name = name
        self._unique_counters = {}  # type: Dict[str, int]
        self._unique_counters_lock = threading.Lock()
        self._rpc_object_map = {}  # type: Dict[str, Optional[RpcObjectManager]]
        self._rpc_object_map_lock = threading.Lock()

        # Remember in which thread this context was created.
        self._creation_thread = threading.current_thread()

        if config is None:
            # Initialize to an empty configuration.
            config = CfgQmi()

        self._config = config

        # Determine workgroup name from configuration.
        self.workgroup_name = self._config.workgroup

        # Create message router.
        self._message_router = MessageRouter(self.name, self.workgroup_name)

        # Create signal manager.
        self._signal_manager = SignalManager(self)
        self._message_router.set_peer_context_callbacks(None, self._signal_manager.handle_peer_context_removed)

        # Create RPC object to answer queries about this context.
        self._internal_make_rpc_object("$context", _ContextRpcObject)

        # List of callback function to be invoked when the context stops.
        self._stop_handlers: List[Callable[[], None]] = []

        # Set object ID.
        self._oid = qmi.object_registry.register(self)

    @property
    def suppress_version_mismatch_warnings(self) -> bool:
        """If set to `True`, no warnings will be issued when connecting to a
        peer that runs a different version of QMI.
        """

        return self._message_router.suppress_version_mismatch_warnings

    @suppress_version_mismatch_warnings.setter
    def suppress_version_mismatch_warnings(self, value: bool) -> None:

        self._message_router.suppress_version_mismatch_warnings = value

    def __repr__(self) -> str:
        return "QMI_Context(name={!r})".format(self.name)

    def _check_in_context_thread(self) -> None:
        """Check that this function is called in the same thread that created the context.

        Raises:
            QMI_WrongThreadException: If the function is called from any
                thread other than the thread that created this context.
        """
        if threading.current_thread() is not self._creation_thread:
            raise QMI_WrongThreadException("Not in context main thread")

    def get_config(self) -> CfgQmi:
        """Return the QMI configuration.

        The QMI configuration is a hierarchical data structure consisting
        of Python dataclass instances. The QMI configuration is typically
        initialized by reading the QMI configuration file during program startup.

        It is not allowed to modify the configuration data structure
        returned by this function.

        Returns:
            The QMI configuration data.
        """
        return self._config

    def get_context_config(self) -> CfgContext:
        """Return the subset of the QMI configuration specific for this context.

        The QMI configuration contains a mapping that applies to this context.

        Returns:
            The QMI configuration subset specific for this context.
        """
        ctxcfg = self._config.contexts.get(self.name)
        if ctxcfg is None:
            ctxcfg = CfgContext()
        return ctxcfg

    def get_qmi_home_dir(self) -> str:
        """Return the QMI home directory.

        The QMI home directory is the location where QMI expects to find
        configuration files and data.

        The QMI home directory may be specified in the QMI configuration file.
        Otherwise, if the environment variable `QMI_HOME` is set, it contains
        the path of the QMI home directory.
        Otherwise, the user home directory will be used as QMI home directory.

        Returns:
            The QMI home directory.
        """

        # Try to get home directory from configuration.
        cfg = self.get_config()
        qmi_home = cfg.qmi_home

        # Try to get home directory from environment.
        if not qmi_home:
            qmi_home = os.getenv("QMI_HOME")

        # Take user home directory as fall-back.
        if not qmi_home:
            qmi_home = str(pathlib.Path.home())

        return qmi_home

    def get_log_dir(self) -> str:
        """Return the directory where log files should be written."""
        cfg = self.get_config()
        if cfg.log_dir:
            return self.resolve_file_name(cfg.log_dir)
        else:
            # No log_dir configured; use the QMI home directory.
            return self.get_qmi_home_dir()

    @staticmethod
    def _file_name_resolves_keyword(file_name: str, keyword: str) -> bool:
        """Return True if the specified file name uses a specific "${keyword}" substitution."""
        for m in re.finditer(r"\$(?:\$|([_a-z][_a-z0-9]*)|{([_a-z][_a-z0-9]*)})", file_name, re.IGNORECASE):
            refname = m.group(1) or m.group(2)
            if refname == keyword:
                return True
        return False

    def get_datastore_dir(self) -> str:
        """Return the base directory of the DataStore repository."""

        cfg = self.get_config()
        datastore_dir = cfg.datastore
        if not datastore_dir:
            raise QMI_ConfigurationException("Missing field 'datastore' in QMI configuration")

        # Avoid recursive calls to get_datastore_dir().
        if self._file_name_resolves_keyword(datastore_dir, "datastore"):
            raise QMI_ConfigurationException("Recursive reference to 'datastore' directory")

        return self.resolve_file_name(datastore_dir)

    def resolve_file_name(self, file_name: str) -> str:
        """Apply substitutions to a configured file name.

        The following substitutions are supported::

          $$            -> "$"
          ${context}    -> context name
          ${qmi_home}   -> QMI home directory
          ${datastore}  -> QMI datastore directory
          ${config_dir} -> directory of QMI configuration file
          ${date}       -> date of program start, UTC, formatted as YYYY-mm-dd
          ${datetime}   -> time of program start, UTC, formatted as YYYY-mm-ddTHH-MM-SS
        """

        # Fast path.
        if "$" not in file_name:
            return file_name

        qmi_home = self.get_qmi_home_dir()

        # Determine configuration directory.
        config_file = self._config.config_file
        if config_file:
            config_dir = os.path.dirname(config_file)
        else:
            config_dir = qmi_home

        # Format date and time.
        gmtime = time.gmtime(self._start_time)
        date = time.strftime("%Y-%m-%d", gmtime)
        datetime = time.strftime("%Y-%m-%dT%H-%M-%S", gmtime)

        mapping = {
            "context": self.name,
            "qmi_home": qmi_home,
            "config_dir": config_dir,
            "date": date,
            "datetime": datetime
        }

        # Avoid unnecessary calls to get_datastore_dir() because
        # - such calls fail if the datastore path is not configured;
        # - such calls lead to infinite recursion when resolving the datastore path.
        if self._file_name_resolves_keyword(file_name, "datastore"):
            mapping["datastore"] = self.get_datastore_dir()

        return string.Template(file_name).substitute(mapping)

    def start(self) -> None:
        """Start the context.

        A `QMI_Context` instance must be started before creating any tasks
        or sending any messages. This happens automatically when the
        application calls ``qmi.start(...)``.
        """

        self._check_in_context_thread()

        _logger.info("Starting QMI context %r", self.name)
        if self._active:
            raise QMI_UsageException("QMI_Context already started")

        # Check that this context has never been started yet.
        if self._used:
            raise QMI_UsageException("Can not start QMI_Context a second time")

        # Start message router.
        self._message_router.start()

        # Start TCP server if a TCP server port is specified in the configuration.
        ctxcfg = self.get_context_config()
        if ctxcfg.tcp_server_port is not None:
            self._message_router.start_tcp_server(ctxcfg.tcp_server_port)

        # The UDP responder is mandatory.
        self._message_router.start_udp_responder(self.DEFAULT_UDP_RESPONDER_PORT)

        # Mark that we're now active.
        self._active = True
        self._used = True
        _active_context_counter.inc()
        _logger.info("QMI context %r now active", self.name)

    def stop(self) -> None:
        """Stop the context.

        After the context is stopped, it must not be used anymore.
        A stopped context may not be started a second time.
        """

        self._check_in_context_thread()

        _logger.info("Stopping QMI context %r", self.name)
        if not self._active:
            raise QMI_UsageException("QMI_Context already inactive")

        # Invoke callback functions to allow subsystems to stop.
        for stop_handler in self._stop_handlers:
            try:
                stop_handler()
            except Exception as exc:
                # The callback function failed, but we still want to
                # continue shutting down QMI. Report the exception and ignore.
                _logger.exception("QMI stop handler failed: %s", str(exc))
            del stop_handler

        self._message_router.stop()

        with self._rpc_object_map_lock:

            # Mark that we're now inactive.
            # No new RPC objects can be created after this point.
            self._active = False

            # Find remaining RPC objects managed by this context.
            managers = []
            for (rpc_object_name, manager) in list(self._rpc_object_map.items()):
                if manager is not None:
                    managers.append(manager)
                    del self._rpc_object_map[rpc_object_name]

        # Stop remaining RPC objects.
        for manager in managers:
            self.unregister_message_handler(manager)
            manager.stop()

        # Update number of active contexts.
        _active_context_counter.dec()

        qmi.object_registry.unregister(self._oid)

    def shutdown_requested(self) -> bool:
        """Return True if the context has received a shutdown request via RPC.

        A background program may call this function to check whether a shutdown
        request has been received. If this function returns True, the program
        should clean up and exit.
        """
        return self._context_shutdown_requested.is_set()

    def wait_until_shutdown(self, duration: float) -> bool:
        """Wait until a shutdown request is received.

        Parameters:
            duration: Maximum wait time in seconds.

        Returns:
            True if a shutdown request is received, False if the wait
            duration expires before a shutdown request is received.
        """
        return self._context_shutdown_requested.wait(duration)

    def register_stop_handler(self, handler: Callable[[], None]) -> None:
        """Register a callback function to be invoked when this context stops.

        Parameters:
            handler: Function to be called (without arguments) when the context stops.
        """
        self._check_in_context_thread()
        self._stop_handlers.append(handler)

    def send_message(self, message: QMI_Message) -> None:
        """Send a message to its destination.

        If the destination is within the local context, this function delivers
        the message to the local message handler. Otherwise, this function
        sends the message to the correct peer context.

        This method is thread-safe. It can be called from any thread.

        Raises:
            ~qmi.core.exceptions.QMI_MessageDeliveryException: If the message can not be routed.
        """
        self._message_router.send_message(message)

    def connect_to_peer(self,
                        peer_context_name: str,
                        peer_address: Optional[str] = None,
                        ignore_duplicate: bool = False
                        ) -> None:
        """Connect to the specified peer context.

        Parameters:
            peer_context_name: Name of the peer context to connect to.
            peer_address: IP address and TCP port of the peer context, formatted as ``"<addr>:<port>"``.
                When not specified, the peer address will be taken from the QMI configuration.
            ignore_duplicate: When True, nothing will happen if a connection
                to the specified context already exists.
        """
        self._check_in_context_thread()
        if not self._active:
            raise QMI_InvalidOperationException("Inactive context can not connect to peer")

        if ignore_duplicate:
            # Do nothing if we are already connected to this peer context.
            if self.has_peer_context(peer_context_name):
                return

        if not peer_address:
            # Get remote context configuration.
            cfg = self.get_config()
            peer_cfg = cfg.contexts.get(peer_context_name)
            if peer_cfg is None:
                raise QMI_UnknownNameException("Unknown remote context {}".format(peer_context_name))
            if (not peer_cfg.host) or (not peer_cfg.tcp_server_port):
                raise QMI_ConfigurationException("Missing host/port for peer context {}".format(peer_context_name))
            peer_address = format_address_and_port((peer_cfg.host, int(peer_cfg.tcp_server_port)))

        # Make new peer connection.
        self._message_router.connect_to_peer(peer_context_name, peer_address)

    def disconnect_from_peer(self, peer_context_name: str) -> None:
        """Disconnect from the specified remote QMI context.

        Raises:
            ~qmi.core.exceptions.QMI_UnknownNameException: If the specified context is not connected.
        """
        self._check_in_context_thread()
        if not self._active:
            raise QMI_InvalidOperationException("Inactive context can not have peers")
        self._message_router.disconnect_from_peer(peer_context_name)

    def has_peer_context(self, peer_context_name: str) -> bool:
        """Return True if a context with specified name is currently connected as a peer.

        Note that the result of this method may become invalid at any time
        as the set of connected contexts may change asynchronously.
        """
        context_names = self._message_router.get_peer_context_names()
        return peer_context_name in context_names

    def discover_peer_contexts(self,
                               workgroup_name_filter: Optional[str] = None,
                               context_name_filter: str = "*") -> List[Tuple[str, str]]:
        """Discover QMI contexts on the network.

        Returns a list of (context_name, address:port)-pairs that can be passed to `connect_to_peer`.

        You can filter on workgroup name and/or context name via the optional arguments. Use a "*" to match any
        sequence of characters (e.g. "ba*" matches "bar", "baz" and "ball") and a "?" to match a single character (e.g.
        "ba?" matches "bar" and "baz", but not "ball"). Filters are case-sensitive.

        Parameters:
            workgroup_name_filter:  filter on workgroup name (None: use the workgroup name of this context's config).
            context_name_filter:    filter on context name (default: "*").
        """
        if workgroup_name_filter is None:
            workgroup_name_filter = self._config.workgroup

        responses = ping_qmi_contexts(
            workgroup_name_filter=workgroup_name_filter,
            context_name_filter=context_name_filter
        )
        contexts: List[Tuple[str, str]] = []
        for resp in responses:
            address = resp.incoming_address[0]
            port = resp.response_packet.context.port
            context_name = resp.response_packet.context.name.decode()

            if context_name != self.name:
                contexts.append((context_name, f"{address}:{port}"))

        return contexts

    def register_message_handler(self, message_handler: QMI_MessageHandler) -> None:
        """Register a new message handler.

        The message handler must be a local object (in the same context)
        and it must have a unique object ID within the context.

        This function is intended for internal use within QMI.
        Application programs should not need to call this function.

        Parameters:
            message_handler: New message handler to be registered.
        """
        self._message_router.register_message_handler(message_handler)

    def unregister_message_handler(self, message_handler: QMI_MessageHandler) -> None:
        """Unregister an existing message handler.

        This function removes a message handler from the context.
        After this function returns, no further messages will be delivered
        to the specified handler.
        The specified object must currently be registered as a message handler.

        This function is intended for internal use within QMI.
        Application programs should not need to call this function.

        Parameters:
            message_handler: The message handler to be unregistered.
        """
        self._message_router.unregister_message_handler(message_handler)

    def info(self) -> str:
        """Return information about the context."""

        ret = []
        ret.append("*** QMI_Context info: name={!r}, workgroup={!r}, pid={!r} ***".format(
            self.name, self.workgroup_name, os.getpid()))
        ret.append("")
        ret.append("*** end of QMI_Context info ***")
        return "\n".join(ret)

    def get_version(self) -> str:
        """Return the QMI version used by this context."""
        return qmi.__version__

    def get_tcp_server_port(self) -> int:
        """Return the TCP server port for this context, or return 0 if this context does not have a TCP server."""
        if self._active:
            return self._message_router.tcp_server_port
        else:
            return 0

    def get_rpc_object_descriptors(self) -> List[RpcObjectDescriptor]:
        """Return a list of descriptors for all local RPC objects."""
        ret = []
        with self._rpc_object_map_lock:
            for manager in self._rpc_object_map.values():
                if manager is not None:
                    rpc_object = manager.rpc_object()
                    ret.append(rpc_object.rpc_object_descriptor)
        return ret

    def get_rpc_object_descriptor(self, rpc_object_name: str) -> Optional[RpcObjectDescriptor]:
        """Return a descriptor for the specified local RPC object."""
        with self._rpc_object_map_lock:
            manager = self._rpc_object_map.get(rpc_object_name)
            if manager is not None:
                rpc_object = manager.rpc_object()
                return rpc_object.rpc_object_descriptor
        return None

    def make_unique_address(self, prefix: str) -> QMI_MessageHandlerAddress:
        """Generate a unique message handler address with a specified prefix."""
        with self._unique_counters_lock:
            nr = self._unique_counters.get(prefix, 0) + 1
            self._unique_counters[prefix] = nr
        return QMI_MessageHandlerAddress(self.name, prefix + str(nr))

    def make_unique_token(self, prefix: str = "$lock_") -> QMI_LockTokenDescriptor:
        """Generate a unique token descriptor."""
        with self._unique_counters_lock:
            nr = self._unique_counters.get(prefix, 0) + 1
            self._unique_counters[prefix] = nr
        return QMI_LockTokenDescriptor(self.name, prefix + str(nr))

    def make_rpc_object(self,
                        rpc_object_name: str,
                        rpc_object_class: Type[QMI_RpcObject],
                        *args: Any,
                        **kwargs: Any
                        ) -> Any:
        """Create an instance of a `QMI_RpcObject` and return a proxy for the new object instance.

        The actual object instance will be created in a separate background thread.
        You can call its methods via RPC, using the proxy object returned by this function.

        Parameters:
            rpc_object_name: Name for the new object instance, unique within the local context.
            rpc_object_class: Class that implements this object (must be a subclass of `QMI_RpcObject`).
            args: Optional arguments for the class constructor.

        Returns:
            An RPC proxy that provides access to the new object instance.
        """

        # The return type of the make_XXX() methods is not annotated.
        # It is not possible to provide a precise annotation, because
        # the actual return type is "rpc_object_class.Proxy" which is
        # a programmatically constructed class and thus not available
        # for static type checking.

        if not is_valid_object_name(rpc_object_name):
            raise QMI_UsageException("Invalid object name {!r}".format(rpc_object_name))

        return self._internal_make_rpc_object(rpc_object_name, rpc_object_class, *args, **kwargs)

    def _internal_make_rpc_object(self,
                                  rpc_object_name: str,
                                  rpc_object_class: Type[QMI_RpcObject],
                                  *args: Any,
                                  **kwargs: Any,
                                  ) -> QMI_RpcProxy:

        # Helper function to create the actual RPC object instance.
        def rpc_object_maker() -> QMI_RpcObject:
            return rpc_object_class(self, rpc_object_name, *args, **kwargs)

        manager = None
        proxy = None

        with self._rpc_object_map_lock:

            # Check that context is active.
            # Only internal RPC objects may be created when the context is not active.
            if (not self._active) and (not rpc_object_name.startswith("$")):
                raise QMI_InvalidOperationException("Can not create RPC object in inactive context")

            # Check unique name.
            if rpc_object_name in self._rpc_object_map:
                raise QMI_DuplicateNameException("Duplicate object name {}".format(rpc_object_name))

            # Reserve name.
            self._rpc_object_map[rpc_object_name] = None

        try:
            # Create RPC object manager and start it.
            address = QMI_MessageHandlerAddress(self.name, rpc_object_name)
            manager = RpcObjectManager(address, self, rpc_object_maker)
            manager.start()

            try:
                # Make a proxy for the RPC object.
                # This blocks until the RPC object is initialized,
                # and may raise an exception if initialization fails.
                proxy = manager.make_proxy()

                with self._rpc_object_map_lock:

                    # Check that context is still active.
                    if (not self._active) and (not rpc_object_name.startswith("$")):
                        proxy = None
                        raise QMI_InvalidOperationException("Can not create RPC object in inactive context")

                    # Register the object manager under the claimed name.
                    self._rpc_object_map[rpc_object_name] = manager

                # Register the object manager as message handler.
                self.register_message_handler(manager)

            finally:
                if proxy is None:
                    # Initialization failed. Shut down the object manager.
                    manager.stop()
                    manager = None

        finally:
            if manager is None:
                # RPC object creation failed. Release the claimed name.
                with self._rpc_object_map_lock:
                    del self._rpc_object_map[rpc_object_name]

        return proxy

    def remove_rpc_object(self, proxy: QMI_RpcProxy) -> None:
        """Stop the specified RPC object and release its resources.

        This can only be used for RPC objects running in the local context.

        Parameters:
            proxy: Proxy for the RPC object to be removed.
        """

        # Check that the proxy refers to a local object.
        if proxy._rpc_object_address.context_id != self.name:
            raise QMI_UsageException("Can not remove remote RPC object {}".format(proxy._rpc_object_address))

        rpc_object_name = proxy._rpc_object_address.object_id

        with self._rpc_object_map_lock:

            # Find the RPC object manager.
            manager = self._rpc_object_map.get(rpc_object_name)
            if manager is None:
                raise QMI_UnknownNameException("Can not remove unknown RPC object {}".format(rpc_object_name))

            # Mark the object as being removed.
            self._rpc_object_map[rpc_object_name] = None

        # Unregister the manager as message handler.
        self.unregister_message_handler(manager)

        # Drop subscriptions on signals published by this object.
        self._signal_manager.handle_object_removed(rpc_object_name)

        # Release the claimed object name.
        with self._rpc_object_map_lock:
            del self._rpc_object_map[rpc_object_name]

        # Shut down the object manager.
        manager.stop()

    def make_instrument(self,
                        instrument_name: str,
                        instrument_class: Type[QMI_Instrument],
                        *args: Any,
                        **kwargs: Any
                        ) -> Any:
        """Create an instance of a `QMI_Instrument` subclass and make it accessible via RPC.

        The actual instrument instance will be created in a separate background thread.
        To access the instrument, you can call its methods via RPC.

        Parameters:
            instrument_name: Unique name for the new instrument instance.
                This name will also be used to access the instrument via RPC.
            instrument_class: Class that implements this instrument (must be a subclass of `QMI_Instrument`).
            args: Optional arguments for the instrument class constructor.

        Returns:
            An RPC proxy that provides access to the new instrument instance.
        """
        return self.make_rpc_object(instrument_name, instrument_class, *args, **kwargs)

    def make_task(self,
                  task_name: str,
                  task_class: Type[QMI_Task],
                  *args: Any,
                  task_runner: Type[QMI_TaskRunner] = QMI_TaskRunner,
                  **kwargs: Any
                  ) -> Any:
        """Create an instance of a `QMI_Task` subclass and make it accessible via RPC.

        The task instance will be created in a separate thread.
        A `QMI_TaskRunner` will be created to manage the task thread.
        The task can be accessed by calling the methods of the task runner via RPC.

        Note that the task is not yet started.
        To start the task, perform an explicit call to the `start()` method of the returned task.

        Parameters:
            task_name: Unique name for the new task instance.
                This name will also be used to access the task runner via RPC.
            task_class: Class that implements this task (must be a subclass of `QMI_Task`).
            task_runner: Class that implements the managing of the task (must be a subclass of `QMI_Taskrunner`)
            args: Optional arguments for the task class constructor.

        Returns:
            An RPC proxy that provides access to the new task.
        """
        return self.make_rpc_object(task_name, task_runner, task_class, args, kwargs)

    def make_proxy(self, rpc_object_descriptor: RpcObjectDescriptor) -> Any:
        """Return a proxy for the specified object descriptor.

        This function is intended for internal use within QMI. Application programs
        should not call this function directly.

        Parameters:
            rpc_object_descriptor: `RpcObjectDescriptor` uniquely identifying an RPC object.

        Returns:
            Proxy for the specified RPC object.
        """
        assert isinstance(rpc_object_descriptor, RpcObjectDescriptor)
        return QMI_RpcProxy(self, rpc_object_descriptor)

    def _make_peer_context_descriptor(self, context_name: str) -> RpcObjectDescriptor:
        address = QMI_MessageHandlerAddress(context_id=context_name, object_id="$context")
        category = _ContextRpcObject.get_category()
        interface = make_interface_descriptor(_ContextRpcObject)
        return RpcObjectDescriptor(address=address, category=category, interface=interface)

    def make_peer_context_proxy(self, context_name: str) -> Any:
        """Return a proxy for the internal `$context` object of the specified peer context.

        This function is intended for internal use within QMI. Application programs
        should not call this function directly.

        Parameters:
            context_name: Name of the peer context.

        Returns:
            A proxy for the internal context RPC object of the specified peer context.
        """
        descriptor = self._make_peer_context_descriptor(context_name)
        return self.make_proxy(descriptor)

    def _get_context_descriptors(self, include_self: bool = True) -> List[RpcObjectDescriptor]:
        context_names = []
        if include_self:
            context_names.append(self.name)
        context_names.extend(self._message_router.get_peer_context_names())

        return [
            self._make_peer_context_descriptor(context_name) for context_name in context_names
            if not context_name.startswith("$")
        ]

    def show_network_contexts(self):
        """Show a list of contexts that are on the network."""
        contexts = self.discover_peer_contexts()

        print()
        if len(contexts) == 0:
            print("No contexts found.")
        else:
            all_names, all_addresses = zip(*contexts)
            max_name_length = max(len(name) for name in all_names)
            max_addr_length = max(len(addr) for addr in all_addresses)

            name_hdr = "name"
            addr_hdr = "address"
            conn_hdr = "connected"

            name_wdt = max(len(name_hdr), max_name_length)
            addr_wdt = max(len(addr_hdr), max_addr_length)
            conn_wdt = len(conn_hdr)

            print("{} {} {}".format(name_hdr.ljust(name_wdt), addr_hdr.ljust(addr_wdt), conn_hdr))
            print("{} {} {}".format("-" * name_wdt, "-" * addr_wdt, "-" * conn_wdt))
            for context_name, context_address in contexts:
                if self.has_peer_context(context_name):
                    connected = "yes"
                else:
                    connected = "no"

                print("{} {} {}".format(context_name.ljust(name_wdt), context_address.ljust(addr_wdt), connected))

            print("{} {} {}".format("-" * name_wdt, "-" * addr_wdt, "-" * conn_wdt))

    def list_rpc_objects(self, category: Optional[str] = None) -> List[Tuple[str, str]]:
        """Returns a list of tuple strings of RPC objects addresses in the local context and in peer context."""
        rpc_objects = []
        for peer_context_descriptor in self._get_context_descriptors():
            peer_context_proxy = self.make_proxy(peer_context_descriptor)
            rpc_object_descriptors = peer_context_proxy.get_rpc_object_descriptors()
            for rpc_object_descriptor in rpc_object_descriptors:
                if category is not None and rpc_object_descriptor.category != category:
                    continue

                rpc_object = (str(rpc_object_descriptor.address), str(rpc_object_descriptor.interface.rpc_class_name))
                rpc_objects.append(rpc_object)

        rpc_objects.sort()
        return rpc_objects

    def show_rpc_objects(self, category: Optional[str] = None) -> None:
        """Show a list of RPC objects in the local context and peer contexts."""

        rpc_objects = self.list_rpc_objects(category)

        print()
        if len(rpc_objects) == 0:
            print("No objects found.")

        else:
            max_address_length = max(max(len(address) for address, _ in rpc_objects), 7)
            max_class_name_length = max(max(len(class_name) for _, class_name in rpc_objects), 4)

            print("{}  {}".format("address".ljust(max_address_length), "type"))
            print("{}  {}".format("-" * max_address_length, "-" * max_class_name_length))
            for address, class_name in rpc_objects:
                print("{}  {}".format(address.ljust(max_address_length), class_name))
            print("{}  {}".format("-" * max_address_length, "-" * max_class_name_length))

    def show_tasks(self) -> None:
        """Show a list of tasks in the local context and peer contexts."""
        self.show_rpc_objects(QMI_TaskRunner.get_category())

    def show_instruments(self) -> None:
        """Show a list of instruments in the local context and peer contexts."""
        self.show_rpc_objects(QMI_Instrument.get_category())

    def show_contexts(self) -> None:
        """Show a list of currently connected contexts (including the local context)."""
        self.show_rpc_objects(_ContextRpcObject.get_category())

    def get_rpc_object_by_name(self, rpc_object_name: str, auto_connect: bool = False,
                               host_port: Optional[str] = None) -> Any:
        """Return a proxy for the specified RPC object.

        The object may exist either in the local context, or in a peer context. Note that when using auto_connect
        keyword parameter, in Windows environment the peer might not be found without giving the exact host:port
        address for `connect_to_peer` command.

        Parameters:
            rpc_object_name: Object name, formatted as ``"<context_name>.<object_id>"``.
            auto_connect: if True, connect automatically to the RPC object peer.
            host_port: Optional host:port string pattern to guide the auto_connect.

        Returns:
            A proxy for the specified object.
        """
        if auto_connect:
            self.connect_to_peer(rpc_object_name.split(".")[0], peer_address=host_port, ignore_duplicate=True)

        (context_id, object_id) = rpc_object_name.split(".", 1)
        context_proxy = self.make_peer_context_proxy(context_id)
        rpc_object_descriptor = context_proxy.get_rpc_object_descriptor(object_id)
        if rpc_object_descriptor is not None:
            assert (rpc_object_descriptor.address.context_id == context_id) and \
                   (rpc_object_descriptor.address.object_id == object_id)
            return self.make_proxy(rpc_object_descriptor)
        raise ValueError("Unknown RPC object '{}'.".format(rpc_object_name))

    def get_instrument(self, instrument_name: str, auto_connect: bool = False, host_port: Optional[str] = None) -> Any:
        """Return a proxy for the specified instrument.

        The instrument may exist either in the local context, or in a peer context.

        Parameters:
            instrument_name: Instrument name, formatted as ``"<context_name>.<instrument_name>"``.
            auto_connect: if True, connect automatically to the instrument peer.
            host_port: Optional host:port string pattern to guide the auto_connect.

        Returns:
            A proxy for the specified instrument.
        """
        return self.get_rpc_object_by_name(instrument_name, auto_connect=auto_connect, host_port=host_port)

    def get_task(self, task_name: str, auto_connect: bool = False, host_port: Optional[str] = None) -> Any:
        """Return a proxy for the specified task.

        The task may exist either in the local context, or in a peer context.

        Parameters:
            task_name: Task name, formatted as ``"<context_name>.<task_name>"``.
            auto_connect: if True, connect automatically to the task peer.
            host_port: Optional host:port string pattern to guide the auto_connect.

        Returns:
            A proxy for the specified task.
        """
        return self.get_rpc_object_by_name(task_name, auto_connect=auto_connect, host_port=host_port)

    def get_configured_contexts(self) -> Dict[str, CfgContext]:
        """ Get active QMI contexts.

        Returns:
            An OrderedDict object of contexts in the config file.
        """
        return self._config.contexts

    def subscribe_signal(self,
                         publisher_context: str,
                         publisher_name: str,
                         signal_name: str,
                         receiver: QMI_SignalReceiver) -> None:
        """Subscribe to a specified signal.

        While subscribed, the SignalReceiver object will receive and queue
        all published signals of the specified type.

        A SignalReceiver object can be simultaneously subscribed to multiple
        signals (from different publishers). Similarly, multiple receivers
        can be simultaneously subscribed to the same signal. However it
        is an error to try to subscribe a receiver to a signal to which it
        is already subscribed.

        This method may safely be called from any thread.

        Parameters:
            publisher_context: Name of the context that publishes the signal.
                An empty string may be used to refer to the local context.
            publisher_name: Name of the publisher of the signal (e.g. instrument name).
            signal_name: Name of the signal to subscribe to.
            receiver: A SignalReceiver object which will receive the published signals.
        """
        self._signal_manager.subscribe_signal(publisher_context, publisher_name, signal_name, receiver)

    def unsubscribe_signal(self,
                           publisher_context: str,
                           publisher_name: str,
                           signal_name: str,
                           receiver: QMI_SignalReceiver) -> None:
        """Unsubscribe from a specified signal.

        It is an error to unsubscribe a receiver from a signal to which is not currently subscribed.

        This method may safely be called from any thread.

        Parameters:
            publisher_context: Name of the context that publishes the signal, or empty string for local context.
            publisher_name: Name of the publisher of the signal.
            signal_name: Name of the signal to unsubscribe from.
            receiver: The SignalReceiver object to unsubscribe from the signal.
        """
        self._signal_manager.unsubscribe_signal(publisher_context, publisher_name, signal_name, receiver)

    def publish_signal(self, publisher_name: str, signal_name: str, *args: Any) -> None:
        """Publish the specified signal to the QMI network.

        The published signal will be received by any SignalReceivers that are
        currently subscribed to the specified signal from the specified publisher.

        Parameters:
            publisher_name: Name of the publisher of the signal.
            signal_name: Name of the signal.
            args: Additional data to send along with the signal.
        """
        self._signal_manager.publish_signal(publisher_name, signal_name, args)



"""Unit test for Wieserlabs FlexDDS NG Dual instrument driver."""

from typing import cast
import unittest
from unittest.mock import MagicMock, call, patch

import qmi
from qmi.core.exceptions import QMI_InstrumentException, QMI_TimeoutException
from qmi.core.transport import QMI_SerialTransport
from qmi.instruments.wieserlabs import OutputChannel, DdsRegister, DcpRegister, Wieserlabs_FlexDdsNg


class TestFlexDDS(unittest.TestCase):

    def setUp(self):
        qmi.start("Test_flex_dds")
        self._transport_mock = MagicMock(spec=QMI_SerialTransport)
        with patch(
                'qmi.instruments.wieserlabs.flexdds.create_transport',
                return_value=self._transport_mock):
            self.instr: Wieserlabs_FlexDdsNg = qmi.make_instrument("instr", Wieserlabs_FlexDdsNg, "transp")
            self.instr = cast(QMI_SerialTransport, self.instr)

    def tearDown(self):
        qmi.stop()

    def _helper_open(self):
        """Open the instrument and check transport interaction."""
        self._transport_mock.read_until.return_value = b"Interactive off\r\n"
        self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self._transport_mock.open.reset_mock()
        self._transport_mock.discard_read.assert_called_once_with()
        self._transport_mock.discard_read.reset_mock()
        self._transport_mock.write.assert_has_calls([
            call(b"\r"),
            call(b"interactive off\r")
        ])
        self._transport_mock.write.reset_mock()
        self._transport_mock.read_until.assert_called_once_with(
            b"Interactive off\r\n", timeout=Wieserlabs_FlexDdsNg.COMMAND_RESPONSE_TIMEOUT)
        self._transport_mock.read_until.reset_mock()

    def test_open_close(self):
        self._helper_open()
        self.instr.close()
        self._transport_mock.close.assert_called_once_with()

    def test_open_fail_transport(self):
        self._transport_mock.open.side_effect = OSError("failed")
        with self.assertRaises(OSError):
            self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self._transport_mock.close.assert_not_called()

    def test_open_fail_protocol(self):
        self._transport_mock.read_until.side_effect = QMI_TimeoutException("failed")
        with self.assertRaises(QMI_TimeoutException):
            self.instr.open()
        self._transport_mock.open.assert_called_once_with()
        self._transport_mock.close.assert_called_once_with()

    def test_get_version(self):
        self._helper_open()
        self._transport_mock.read_until.side_effect = [
            b'Version: "dummy,123" (ignored)\r\n',
            b"USER:\r\n"
        ]
        version = self.instr.get_version()
        self.instr.close()
        self._transport_mock.discard_read.assert_called_once_with()
        self._transport_mock.write.assert_called_once_with(b"version\r")
        self.assertEqual(version, "dummy,123")

    def test_get_version_fail(self):
        self._helper_open()
        self._transport_mock.read_until.side_effect = [
            b'Bad response\r\n',
            b"USER:\r\n"
        ]
        with self.assertRaises(QMI_InstrumentException):
            version = self.instr.get_version()
        self.instr.close()

    def test_get_pll_status(self):
        self._helper_open()
        self._transport_mock.read_until.side_effect = [
            b"LMK PLL status: rv=666\n",
            b"  DLD (lock detect): PLL1: *UNLOCK*   PLL2: locked      BOTH: *UNLOCK*\n",
            b"  Holdover . . . . : *ACTIVE*\n",
            b"  DAC. . . . . . . : low\n",
            b"  LOS (signal loss): CLKIN0: *LOST*       CLKIN1: not lost\n",
            b"  CLKIN selected . : CLKIN0: *SELECT*     CLKIN1: (unsel)\n"
        ]
        status = self.instr.get_pll_status()
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"lmkpll\r")
        self.assertFalse(status.pll1_lock)
        self.assertTrue(status.pll2_lock)
        self.assertTrue(status.holdover)
        self.assertTrue(status.clkin0_lost)
        self.assertFalse(status.clkin1_lost)

    def test_dds_reset(self):
        self._helper_open()
        self._transport_mock.read_until.return_value = b"DDS reset OK\r\n"
        self.instr.dds_reset(OutputChannel.OUT0)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"dds 0 reset\r")

    def test_dcp_start(self):
        self._helper_open()
        self.instr.dcp_start(OutputChannel.OUT1)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"dcp 1 start\r")

    def test_dcp_stop(self):
        self._helper_open()
        self.instr.dcp_stop(OutputChannel.BOTH)
        self.instr.close()
        self._transport_mock.write.assert_called_once_with(b"dcp  stop\r")

    def test_dcp_spi_write(self):
        self._helper_open()
        self.instr.dcp_spi_write(OutputChannel.OUT0, DdsRegister.STP0, 0x1234, wait_spi=False, flush=True)
        self.instr.dcp_spi_write(OutputChannel.OUT1, DdsRegister.CFR1, 0x2345, wait_spi=True)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 0 spi:STP0=0x1234:c!\r"),
            call(b"dcp 1 spi:CFR1=0x2345:w\r")
        ])

    def test_dcp_register_write(self):
        self._helper_open()
        self.instr.dcp_register_write(OutputChannel.OUT0, DcpRegister.AM_S0, 0x111)
        self.instr.dcp_register_write(OutputChannel.BOTH, DcpRegister.CFG_BNC_A, 0x321, flush=True)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 0 wr:AM_S0=0x111\r"),
            call(b"dcp  wr:CFG_BNC_A=0x321!\r")
        ])

    def test_dcp_update(self):
        self._helper_open()
        self.instr.dcp_update(OutputChannel.OUT1, flush=True)
        self.instr.dcp_update(OutputChannel.BOTH)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 1 update:u!\r"),
            call(b"dcp  update:u\r")
        ])

    def test_single_tone(self):
        self._helper_open()
        self.instr.set_single_tone(OutputChannel.OUT0, 100.0e6, 0.9, 0.1)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 0 spi:CFR1=0x410002:w\r"),
            call(b"dcp 0 spi:CFR2=0x14008c0:w\r"),
            call(b"dcp 0 spi:STP0=0x3999199a1999999a:w\r"),
            call(b"dcp 0 update:u\r"),
            call(b"dcp 0 start\r")
        ])

    def test_single_tone_badparam(self):
        self._helper_open()
        with self.assertRaises(ValueError):
            self.instr.set_single_tone(OutputChannel.OUT0, 800.0e6, 0.9, 0.1)
        with self.assertRaises(ValueError):
            self.instr.set_single_tone(OutputChannel.OUT0, 100.0e6, 1.1, 0.1)
        with self.assertRaises(ValueError):
            self.instr.set_single_tone(OutputChannel.OUT0, 100.0e6, 0.9, -0.1)
        self.instr.close()

    def test_amplitude_modulation(self):
        self._helper_open()
        self.instr.set_amplitude_modulation(OutputChannel.OUT0, 10.0e6, 0.3, 0.25, 0, 0.1, 1.1)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 0 spi:CFR1=0x410002:w\r"),
            call(b"dcp 0 spi:CFR2=0x14008d0:w\r"),
            call(b"dcp 0 spi:STP0=0x4000028f5c29:w\r"),
            call(b"dcp 0 wr:AM_O=0x4ccc\r"),
            call(b"dcp 0 wr:AM_S1=0x0\r"),
            call(b"dcp 0 wr:AM_O0=0x199a\r"),
            call(b"dcp 0 wr:AM_S0=0x119a\r"),
            call(b"dcp 0 wr:AM_CFG=0x20000000\r"),
            call(b"dcp 0 update:u\r"),
            call(b"dcp 0 start\r")
        ])

    def test_amplitude_modulation1(self):
        self._helper_open()
        self.instr.set_amplitude_modulation(OutputChannel.OUT1, 10.0e6, 0.3, 0.25, 1, 0.1, 1.1)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 1 spi:CFR1=0x410002:w\r"),
            call(b"dcp 1 spi:CFR2=0x14008d0:w\r"),
            call(b"dcp 1 spi:STP0=0x4000028f5c29:w\r"),
            call(b"dcp 1 wr:AM_O=0x4ccc\r"),
            call(b"dcp 1 wr:AM_S0=0x0\r"),
            call(b"dcp 1 wr:AM_O1=0x199a\r"),
            call(b"dcp 1 wr:AM_S1=0x119a\r"),
            call(b"dcp 1 wr:AM_CFG=0x20000000\r"),
            call(b"dcp 1 update:u\r"),
            call(b"dcp 1 start\r")
        ])

    def test_amplitude_modulation_badparam(self):
        self._helper_open()
        with self.assertRaises(ValueError):
            self.instr.set_amplitude_modulation(OutputChannel.OUT1, 800.0e6, 0.3, 0.25, 0, 0.1, 1.1)
        with self.assertRaises(ValueError):
            self.instr.set_amplitude_modulation(OutputChannel.OUT1, 10.0e6, 2.1, 0.25, 0, 0.1, 1.1)
        with self.assertRaises(ValueError):
            self.instr.set_amplitude_modulation(OutputChannel.OUT1, 10.0e6, 0.3, -0.1, 0, 0.1, 1.1)
        with self.assertRaises(ValueError):
            self.instr.set_amplitude_modulation(OutputChannel.OUT1, 10.0e6, 0.3, 0.25, 3, 0.1, 1.1)
        with self.assertRaises(ValueError):
            self.instr.set_amplitude_modulation(OutputChannel.OUT1, 10.0e6, 0.3, 0.25, 0, 0.6, 1.1)
        with self.assertRaises(ValueError):
            self.instr.set_amplitude_modulation(OutputChannel.OUT1, 10.0e6, 0.3, 0.25, 0, 0.1, 31.0)
        self.instr.close()

    def test_frequency_modulation(self):
        self._helper_open()
        self.instr.set_frequency_modulation(OutputChannel.OUT0, 50.0e6, 0.7, 0.0, 1, -0.1, 5.0e6)
        self.instr.close()
        # fm_gain = 9
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 0 spi:CFR1=0x410002:w\r"),
            call(b"dcp 0 spi:CFR2=0x14008d9:w\r"),
            call(b"dcp 0 spi:STP0=0x2ccc00000bcccccd:w\r"),
            call(b"dcp 0 spi:FTW=0xbcccccd:w\r"),
            call(b"dcp 0 wr:AM_O=0x8000\r"),
            call(b"dcp 0 wr:AM_S0=0x0\r"),
            call(b"dcp 0 wr:AM_O1=0x3e666\r"),
            call(b"dcp 0 wr:AM_S1=0xa3d\r"),
            call(b"dcp 0 wr:AM_CFG=0x20000002\r"),
            call(b"dcp 0 update:u\r"),
            call(b"dcp 0 start\r")
        ])

    def test_frequency_modulation1(self):
        self._helper_open()
        self.instr.set_frequency_modulation(OutputChannel.OUT1, 3.0e6, 0.7, 0.0, 0, -0.1, 5.0e6)
        self.instr.close()
        # fm_gain = 9
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 1 spi:CFR1=0x410002:w\r"),
            call(b"dcp 1 spi:CFR2=0x14008d9:w\r"),
            call(b"dcp 1 spi:STP0=0x2ccc0000000001a6:w\r"),
            call(b"dcp 1 spi:FTW=0x1a6:w\r"),
            call(b"dcp 1 wr:AM_O=0x624d\r"),
            call(b"dcp 1 wr:AM_S1=0x0\r"),
            call(b"dcp 1 wr:AM_O0=0x3e666\r"),
            call(b"dcp 1 wr:AM_S0=0xa3d\r"),
            call(b"dcp 1 wr:AM_CFG=0x20000002\r"),
            call(b"dcp 1 update:u\r"),
            call(b"dcp 1 start\r")
        ])

    def test_frequency_modulation_badparam(self):
        self._helper_open()
        with self.assertRaises(ValueError):
            self.instr.set_frequency_modulation(OutputChannel.OUT0, 500.0e6, 0.7, 0.0, 1, -0.1, 5.0e6)
        with self.assertRaises(ValueError):
            self.instr.set_frequency_modulation(OutputChannel.OUT0, 50.0e6, 1.1, 0.0, 1, -0.1, 5.0e6)
        with self.assertRaises(ValueError):
            self.instr.set_frequency_modulation(OutputChannel.OUT0, 50.0e6, 0.7, 1.1, 1, -0.1, 5.0e6)
        with self.assertRaises(ValueError):
            self.instr.set_frequency_modulation(OutputChannel.OUT0, 50.0e6, 0.7, 0.0, 3, -0.1, 5.0e6)
        with self.assertRaises(ValueError):
            self.instr.set_frequency_modulation(OutputChannel.OUT0, 50.0e6, 0.7, 0.0, 1, -0.6, 5.0e6)
        with self.assertRaises(ValueError):
            self.instr.set_frequency_modulation(OutputChannel.OUT0, 50.0e6, 0.7, 0.0, 1, -0.1, 600.0e6)
        self.instr.close()

    def test_digital_modulation(self):
        self._helper_open()
        self.instr.set_digital_modulation(OutputChannel.OUT0, 400.0e6, 0.7, 0.0, 1, False)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 0 wr:CFG_BNC_B=0x0\r"),
            call(b"dcp 0 wr:CFG_OSK=0x408\r"),
            call(b"dcp 0 spi:CFR1=0xc10202:w\r"),
            call(b"dcp 0 spi:CFR2=0x14008c0:w\r"),
            call(b"dcp 0 spi:STP0=0x2ccc000066666666:w\r"),
            call(b"dcp 0 spi:ASF=0xb330:w\r"),
            call(b"dcp 0 update:u\r"),
            call(b"dcp 0 start\r")
        ])

    def test_phase_modulation(self):
        self._helper_open()
        self.instr.set_phase_modulation(OutputChannel.OUT0, 10.0e6, 0.7, 0, 0.1, 1.1)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 0 spi:CFR1=0x410002:w\r"),
            call(b"dcp 0 spi:CFR2=0x14008d0:w\r"),
            call(b"dcp 0 spi:STP0=0x2ccc0000028f5c29:w\r"),
            call(b"dcp 0 wr:AM_O=0x0\r"),
            call(b"dcp 0 wr:AM_S1=0x0\r"),
            call(b"dcp 0 wr:AM_O0=0x199a\r"),
            call(b"dcp 0 wr:AM_S0=0x119a\r"),
            call(b"dcp 0 wr:AM_CFG=0x20000001\r"),
            call(b"dcp 0 update:u\r"),
            call(b"dcp 0 start\r")
        ])

    def test_phase_modulation1(self):
        self._helper_open()
        self.instr.set_phase_modulation(OutputChannel.OUT1, 10.0e6, 0.7, 1, 0.1, 1.1)
        self.instr.close()
        self._transport_mock.write.assert_has_calls([
            call(b"dcp 1 spi:CFR1=0x410002:w\r"),
            call(b"dcp 1 spi:CFR2=0x14008d0:w\r"),
            call(b"dcp 1 spi:STP0=0x2ccc0000028f5c29:w\r"),
            call(b"dcp 1 wr:AM_O=0x0\r"),
            call(b"dcp 1 wr:AM_S0=0x0\r"),
            call(b"dcp 1 wr:AM_O1=0x199a\r"),
            call(b"dcp 1 wr:AM_S1=0x119a\r"),
            call(b"dcp 1 wr:AM_CFG=0x20000001\r"),
            call(b"dcp 1 update:u\r"),
            call(b"dcp 1 start\r")
        ])


if __name__ == '__main__':
    unittest.main()



class ps3000a:
    PICO_STATUS = {"PICO_OK": 0}
    PICO_STATUS_LOOKUP = {v: k for k, v in PICO_STATUS.items()}
    PICO_COUPLING = {"AC": 0, "DC": 1}
    PICO_INFO = {"PICO_BATCH_AND_SERIAL": 0x1}
    name = "ps3000a"

    def __init__(self):
        pass

    def ps3000aStop(self):
        return 0

    def ps3000aOpenUnit(self, par_serial):
        return 0

    def ps3000aCloseUnit(self):
        return 0

    def ps3000aGetUnitInfo(self, par_string, max_len, par_size, info):
        return 0

    def ps3000aSetChannel(self, channel, enable, coupling, range_sel, offset):
        return 0

    def ps3000aSetSimpleTrigger(self, enable, channel, threshold, edge, val1, val2):
        return 0

    def ps3000aSetTriggerChannelConditions(self, val1, val2):
        return 0

    def ps3000aIsReady(self, par_ready):
        return 0

    def ps3000aSetDataBuffer(self, chan, buf, buf_len, val1, val2):
        return 0

    def ps3000aGetValues(self, start_index, par_num_samples, down_sample_ratio, down_sample_ratio_mode, segment_index,
                         par_overflow):
        return 0



"""
Instrument driver for the Anapico APSIN Series signal generators.
"""

import logging
import re
import time

from qmi.core.context import QMI_Context
from qmi.core.exceptions import QMI_InstrumentException
from qmi.core.instrument import QMI_Instrument, QMI_InstrumentIdentification
from qmi.core.rpc import rpc_method
from qmi.core.scpi_protocol import ScpiProtocol
from qmi.core.transport import create_transport

# Global variable holding the logger for this module.
_logger = logging.getLogger(__name__)


class Anapico_APSIN(QMI_Instrument):
    """Instrument driver for the Anapico APSIN Series signal generators.

    This driver has been developed for and tested with the Anapico APSIN2010,
    but is expected to also work with other APSIN models.

    A subset of the instrument functionality is supported.
    Continuous wave, pulse modulation, amplitude modulation and frequency modulation are supported.
    Triggering, sweeping and multipurpose output are not supported.
    """

    # Default response timeout in seconds.
    # The instrument responds to most queries in a fraction of a second,
    # however when modulation is enabled, the instrument takes up to 5 seconds
    # to respond to certain commands.
    DEFAULT_RESPONSE_TIMEOUT = 10.0

    def __init__(self,
                 context: QMI_Context,
                 name: str,
                 transport: str
                 ) -> None:
        """Initialize the instrument driver.

        Parameters:
            name: Name for this instrument instance.
            transport: QMI transport descriptor to connect to the instrument.
                       For example "tcp:172.16.xx.yy:18".
        """
        super().__init__(context, name)
        self._transport = create_transport(transport, default_attributes={"port": 18})
        self._scpi_protocol = ScpiProtocol(self._transport, default_timeout=self.DEFAULT_RESPONSE_TIMEOUT)
        self._power_unit_configured = False

    @rpc_method
    def open(self) -> None:
        _logger.info("[%s] Opening connection to instrument", self._name)
        self._check_is_closed()
        self._power_unit_configured = False
        self._transport.open()
        try:
            # Discard pending errors.
            self._scpi_protocol.write("*CLS")
            self._scpi_protocol.ask("SYST:ERR:ALL?")
        except Exception:
            self._transport.close()
            raise
        super().open()

    @rpc_method
    def close(self) -> None:
        _logger.info("[%s] Closing connection to instrument", self._name)
        super().close()
        self._transport.close()

    def _ask(self, cmd: str) -> str:
        """Send a query and return a string response."""
        self._transport.discard_read()  # discard potential stale reply after timeout
        return self._scpi_protocol.ask(cmd)

    def _ask_float(self, cmd: str) -> float:
        """Send a query and return a floating point response."""
        self._transport.discard_read()
        resp = self._scpi_protocol.ask(cmd)
        try:
            return float(resp)
        except ValueError:
            raise QMI_InstrumentException("Unexpected response to command {!r}: {!r}".format(cmd, resp))

    def _ask_bool(self, cmd: str) -> bool:
        """Send a query and return a boolean response."""
        self._transport.discard_read()
        resp = self._scpi_protocol.ask(cmd)
        value = resp.strip().upper()
        if value in ("1", "ON"):
            return True
        elif value in ("0", "OFF"):
            return False
        else:
            raise QMI_InstrumentException("Unexpected response to command {!r}: {!r}".format(cmd, resp))

    def _ask_source_is_ext(self, cmd: str) -> bool:
        """Send a query and return True if the response is "EXT", False if it is "INT"."""
        self._transport.discard_read()
        source = self._scpi_protocol.ask(cmd).strip().upper()
        if source.startswith("EXT"):
            return True
        elif source.startswith("INT"):
            return False
        raise QMI_InstrumentException("Unexpected response {!r} to command {}".format(source, cmd))

    def _check_error(self) -> None:
        """Read the instrument error queue and raise an exception if there is an error."""
        resp = self._ask(":SYST:ERR:ALL?")
        # When there are no errors, the response is '0,"No error"'.
        if not re.match(r"^\s*[-+]?0\s*,", resp):
            # Some error occurred.
            raise QMI_InstrumentException("Instrument returned error: {}".format(resp))

    @rpc_method
    def reset(self) -> None:
        """Reset the instrument, returning most settings to their defaults."""
        self._transport.discard_read()
        self._scpi_protocol.write("*CLS")  # clear error status
        self._scpi_protocol.ask(":SYST:ERR:ALL?")  # clear error message queue
        self._scpi_protocol.write("*RST")  # reset instrument
        time.sleep(0.5)
        self._check_error()

    @rpc_method
    def get_idn(self) -> QMI_InstrumentIdentification:
        """Read instrument type and version and return QMI_InstrumentIdentification instance."""
        resp = self._ask("*IDN?")
        words = resp.rstrip().split(",")
        if len(words) != 4:
            raise QMI_InstrumentException("Unexpected response to *IDN?, got {!r}".format(resp))
        return QMI_InstrumentIdentification(vendor=words[0].strip(),
                                            model=words[1].strip(),
                                            serial=words[2].strip(),
                                            version=words[3].strip())

    @rpc_method
    def get_output_enabled(self) -> bool:
        """Return True if the RF output is enabled, False if disabled."""
        return self._ask_bool(":OUTP?")

    @rpc_method
    def set_output_enabled(self, enable: bool) -> None:
        """Enable or disable the RF output."""
        self._scpi_protocol.write(":OUTP {}".format(1 if enable else 0))
        self._check_error()

    @rpc_method
    def get_frequency(self) -> float:
        """Return the current RF frequency in Hz."""
        return self._ask_float(":FREQ?")

    @rpc_method
    def set_frequency(self, frequency: float) -> None:
        """Set the RF frequency in Hz.

        The instrument supports a limited frequency range.
        If the specified frequency is outside the supported range,
        the instrument will select the closest supported frequency.
        """
        self._scpi_protocol.write(":FREQ {}".format(frequency))
        self._check_error()

    @rpc_method
    def get_phase(self) -> float:
        """Return the current phase adjustment in radians."""
        return self._ask_float(":PHAS?")

    @rpc_method
    def set_phase(self, phase: float) -> None:
        """Set the current phase adjustment in radians."""
        self._scpi_protocol.write(":PHAS {}".format(phase))
        self._check_error()

    def _ensure_power_unit(self) -> None:
        """If the power unit is not yet configured, set it to dBm."""
        if not self._power_unit_configured:
            self._scpi_protocol.write(":UNIT:POW DBM")
            self._check_error()
            self._power_unit_configured = True

    @rpc_method
    def get_power(self) -> float:
        """Return the current RF output power in dBm."""
        self._ensure_power_unit()
        return self._ask_float(":POW?")

    @rpc_method
    def set_power(self, power: float) -> None:
        """Set the RF output power in dBm."""
        self._ensure_power_unit()
        self._scpi_protocol.write(":POW {}".format(power))
        self._check_error()

    @rpc_method
    def get_pulsemod_enabled(self) -> bool:
        """Return True if pulse modulation is enabled, False if disabled."""
        return self._ask_bool(":PULM:STAT?")

    @rpc_method
    def set_pulsemod_enabled(self, enable: bool) -> None:
        """Enable or disable pulse modulation."""
        self._scpi_protocol.write(":PULM:STAT {}".format(1 if enable else 0))
        self._check_error()

    @rpc_method
    def get_pulsemod_ext_source(self) -> bool:
        """Return True if pulse modulation uses an external source, False when using the internal source."""
        return self._ask_source_is_ext(":PULM:SOUR?")

    @rpc_method
    def set_pulsemod_ext_source(self, ext: bool) -> None:
        """Enable or disable the external pulse modulation source.

        When set to True, pulse modulation is controlled by an external
        pulse generator connected to the PULSE port of the instrument.
        When set to False, pulse modulation is controlled by an internal
        pulse generator.
        """
        source = "EXT" if ext else "INT"
        self._scpi_protocol.write(":PULM:SOUR {}".format(source))
        self._check_error()

    @rpc_method
    def get_pulsemod_polarity(self) -> bool:
        """Return True if the external pulse modulation polarity is inverted; False if it is normal."""
        resp = self._ask(":PULM:POL?").strip().upper()
        if resp.startswith("INV"):
            return True
        elif resp.startswith("NORM"):
            return False
        raise QMI_InstrumentException("Unexpected response {!r} ot command :PULM:POL?".format(resp))

    @rpc_method
    def set_pulsemod_polarity(self, inverted: bool) -> None:
        """Enable or disable inverted polarity for external modulation."""
        self._scpi_protocol.write(":PULM:POL {}".format("INV" if inverted else "NORM"))
        self._check_error()

    @rpc_method
    def get_am_enabled(self) -> bool:
        """Return True if amplitude modulation is enabled, False if disabled."""
        return self._ask_bool(":AM:STAT?")

    @rpc_method
    def set_am_enabled(self, enable: bool) -> None:
        """Enable or disable amplitude modulation."""
        self._scpi_protocol.write(":AM:STAT {}".format(1 if enable else 0))
        self._check_error()

    @rpc_method
    def get_am_ext_source(self) -> bool:
        """Return True if amplitude modulation uses an external source, False when using the internal source."""
        return self._ask_source_is_ext(":AM:SOUR?")

    @rpc_method
    def set_am_ext_source(self, ext: bool) -> None:
        """Enable or disable the external amplitude modulation source.

        When set to True, amplitude modulation is controlled by an external
        signal connected to the PULSE port of the instrument. Note the external
        amplitude modulation signal is AC coupled.

        When set to False, amplitude modulation is controlled by an internal signal.
        """
        source = "EXT" if ext else "INT"
        self._scpi_protocol.write(":AM:SOUR {}".format(source))
        self._check_error()

    @rpc_method
    def get_am_sensitivity(self) -> float:
        """Return the amplitude modulation sensitivity as a fraction of nominal amplitude per Volt."""
        return self._ask_float(":AM:SENS?")

    @rpc_method
    def set_am_sensitivity(self, sensitivity: float) -> None:
        """Set the sensitivity for external amplitude modulation.

        Parameters:
            sensitivity: Modulation sensitivity per Volt external input (range 0 ... 3).
        """
        self._scpi_protocol.write(":AM:SENS {}".format(sensitivity))
        self._check_error()

    @rpc_method
    def get_fm_enabled(self) -> bool:
        """Return True if frequency modulation is enabled, False if disabled."""
        return self._ask_bool(":FM:STAT?")

    @rpc_method
    def set_fm_enabled(self, enable: bool) -> None:
        """Enable or disable frequency modulation."""
        self._scpi_protocol.write(":FM:STAT {}".format(1 if enable else 0))
        self._check_error()

    @rpc_method
    def get_fm_ext_source(self) -> bool:
        """Return True if frequency modulation uses an external source, False when using the internal source."""
        return self._ask_source_is_ext(":FM:SOUR?")

    @rpc_method
    def set_fm_ext_source(self, ext: bool) -> None:
        """Enable or disable the external frequency modulation source.

        When set to True, frequency modulation is controlled by an external
        signal connected to the Phi-M port of the instrument.
        When set to False, frequency modulation is controlled by an internal signal.
        """
        source = "EXT" if ext else "INT"
        self._scpi_protocol.write(":FM:SOUR {}".format(source))
        self._check_error()

    @rpc_method
    def get_fm_sensitivity(self) -> float:
        """Return the frequency modulation sensitivity in Hz/Volt."""
        return self._ask_float(":FM:SENS?")

    @rpc_method
    def set_fm_sensitivity(self, sensitivity: float) -> None:
        """Set the sensitivity for external frequency modulation.

        Parameters:
            sensitivity: Frequency modulation deviation per one Volt peak amplitude of external input.
        """
        self._scpi_protocol.write(":FM:SENS {}".format(sensitivity))
        self._check_error()

    @rpc_method
    def get_fm_coupling(self) -> str:
        """Return the input coupling for external frequency modulation.

        Returns:
            Input coupling, either "AC" or "DC".
        """
        return self._ask(":FM:COUP?").strip().upper()

    @rpc_method
    def set_fm_coupling(self, coupling: str) -> None:
        """Select AC or DC coupling for the external frequency modulation input.

        Parameters:
            coupling: Input coupling, either "AC" or "DC".
        """
        coupling = coupling.upper()
        if coupling not in ("AC", "DC"):
            raise ValueError("Unknown value {}".format(coupling))
        self._scpi_protocol.write(":FM:COUP {}".format(coupling))
        self._check_error()

    @rpc_method
    def get_reference_source(self) -> str:
        """Return the current reference clock source.

        Returns:
            "INT" if the internal reference is used;
            "EXT" if the external reference is used.
        """
        return self._ask(":ROSC:SOUR?").strip().upper()

    @rpc_method
    def set_reference_source(self, source: str) -> None:
        """Set the reference clock source.

        Parameters:
            source: "INT" to select the internal reference; "EXT" to select the external reference.
        """
        source = source.upper()
        if source not in ("INT", "EXT"):
            raise ValueError("Unknown value {}".format(source))
        self._scpi_protocol.write(":ROSC:SOUR {}".format(source))
        self._check_error()

    @rpc_method
    def get_external_reference_frequency(self) -> float:
        """Return the configured external reference input frequency in Hz."""
        return self._ask_float(":ROSC:EXT:FREQ?")

    @rpc_method
    def set_external_reference_frequency(self, frequency: float) -> None:
        """Set the expected external reference input frequency in Hz.

        The instrument supports a limited range of reference frequencies.
        If the specified frequency is outside the supported range,
        the instrument will select the closest supported frequency.

        The instrument has a very narrow reference lock range. The external
        reference frequency must match the configured frequency within 1 ppm
        to ensure a proper lock.
        """
        self._scpi_protocol.write(":ROSC:EXT:FREQ {}".format(frequency))
        self._check_error()

    @rpc_method
    def get_reference_is_locked(self) -> bool:
        """Return True if the instrument is locked to the internal or external reference clock."""
        return self._ask_bool(":ROSC:LOCK?")

    @rpc_method
    def get_reference_output_enabled(self) -> bool:
        """Return True if the 10 MHz reference clock output is enabled, False if disabled."""
        return self._ask_bool(":ROSC:OUTP?")

    @rpc_method
    def set_reference_output_enabled(self, enable: bool) -> None:
        """Enable or disable the 10 MHz reference clock output signal."""
        self._scpi_protocol.write(":ROSC:OUTP {}".format(1 if enable else 0))
        self._check_error()



"""
Rohde & Schwarz, analog and vector signal generators.

The qmi.instruments.rohde_schwarz package provides support for:
- SGS100A model
- SMBV100A model.
"""
# Alternative, QMI naming convention approved names
from qmi.instruments.rohde_schwarz.sgs100a import RohdeSchwarz_SGS100A as RohdeSchwarz_Sgs100a
from qmi.instruments.rohde_schwarz.smbv100a import RohdeSchwarz_SMBV100A as RohdeSchwarz_Smbv100a



"""Testcase of Stanford Research Systems Sim900 instrument"""
import logging
import unittest
from unittest.mock import patch, call

import qmi
from qmi.core.exceptions import QMI_InvalidOperationException, QMI_UsageException
from qmi.core.transport import QMI_Transport
from qmi.instruments.stanford_research_systems import Srs_Sim900


class TestSim900(unittest.TestCase):
    """Testcase of Stanford Research Systems Sim900 instrument"""

    def setUp(self):
        qmi.start("test_sim_900")
        # Add patches
        patcher = patch('qmi.instruments.stanford_research_systems.sim900.create_transport', spec=QMI_Transport)
        self.transport_mock = patcher.start().return_value
        self.addCleanup(patcher.stop)
        # Make DUT
        self.sim900 = qmi.make_instrument('Sim900', Srs_Sim900, "")
        self.sim900.open()

    def tearDown(self):
        if self.sim900.is_open():
            self.sim900.close()
        qmi.stop()
        logging.getLogger("qmi.core.instrument").setLevel(logging.NOTSET)

    def test_check_constructor(self):
        # act
        self.sim900.close()
        # assert
        self.transport_mock.open.assert_called()
        self.transport_mock.close.assert_called()

    def test_invalid_use_raises_exception(self):
        # Suppress logging.
        logging.getLogger("qmi.core.instrument").setLevel(logging.CRITICAL)
        self.sim900.close()
        # arrange & act
        with self.assertRaises(QMI_InvalidOperationException):
            self.sim900.input_bytes_waiting(6)

    def test_invalid_port(self):
        error_ports = [0,9]
        # arrange & act
        for error_port in error_ports:
            with self.subTest(i=error_port):
                with self.assertRaises(QMI_UsageException):
                    self.sim900.input_bytes_waiting(error_port)

    def test_input_bytes_waiting(self):
        # arrange
        self.transport_mock.read_until.return_value = b'10\n'
        expected_calls = [
            call.open(),
            call.write(b'NINP? 6\n'),
            call.read_until(message_terminator=b'\n', timeout=0.1)]

        # act
        result = self.sim900.input_bytes_waiting(6)
        # assert
        self.assertEqual(self.transport_mock.method_calls, expected_calls)
        self.assertEqual(result, 10)

    def test_get_raw_bytes(self):
        # arrange
        self.transport_mock.read.return_value = b'0123456789'
        expected_calls = [
            call.open(),
            call.write(b'RAWN? 6,10\n'),
            call.read(10, timeout=0.1)]

        # act
        result = self.sim900.get_raw_bytes(6, 10)
        # assert
        self.assertEqual(self.transport_mock.method_calls, expected_calls)
        self.assertEqual(result, b'0123456789')

    def test_send_terminated_message(self) -> None:
        # arrange
        expected_calls = [
            call.open(),
            call.write(b'SNDT 6,\"hi\"\n')]

        # act
        self.sim900.send_terminated_message(6, 'hi')
        # assert
        self.assertEqual(self.transport_mock.method_calls, expected_calls)

    def test_ask_module(self) -> None:
        # arrange
        self.transport_mock.read_until.return_value = b'10\n'
        self.transport_mock.read.return_value = b'123456789\n'

        expected_calls = [
            call.open(),
            call.write(b'SNDT 6,\"hi\"\n'),
            call.write(b'NINP? 6\n'),
            call.read_until(message_terminator=b'\n', timeout=0.1),
            call.write(b'RAWN? 6,10\n'),
            call.read(10, timeout=0.1)]
        # act
        result = self.sim900.ask_module(6,"hi")

        # assert
        self.assertEqual(self.transport_mock.method_calls, expected_calls)
        self.assertEqual(result, '123456789')



import time
import threading
from typing import cast

import qmi
from qmi.core.rpc import QMI_RpcObject
from qmi.core.pubsub import QMI_Signal


class DummyDataCreator(QMI_RpcObject):

    signal_publisher = QMI_Signal([list])

    def __init__(self, ctx, name):
        super().__init__(ctx, name)
        self._data = [0] * 5
        self._running = False

    def start(self):
        self._running = True
        while self._running:
            self._data.append(self._data.pop(0) + 1)
            self.signal_publisher.publish(self._data)
            time.sleep(1)

    def stop(self):
        self._running = False


if __name__ == "__main__":
    qmi.start("oh_yeah", context_cfg={"oh_yeah": {"tcp_server_port": 12345}})
    # Create an RPC proxy to the class that has the signal publisher.
    ddc = qmi.make_rpc_object("whoa", DummyDataCreator)
    ddc = cast(ddc, DummyDataCreator)(qmi.context(), "whoa")
    qmi.show_rpc_objects()
    pub_thread = threading.Thread(target=ddc.start)
    pub_thread.start()

    while True:
        try:
            time.sleep(1)

        except KeyboardInterrupt:
            break

        except:
            break

    ddc.stop()
    pub_thread.join()
    qmi.stop()
    exit()



"""
Module for the APT protocol used by Thorlabs. The documentation for the protocol can be found
here https://www.thorlabs.com/Software/Motion%20Control/APT_Communications_Protocol.pdf
"""

from ctypes import (
    LittleEndianStructure,
    c_uint8,
    c_uint16,
    c_int16,
    c_uint32,
    c_int32,
    c_char,
    sizeof,
)
from enum import Enum
from typing import ClassVar, Type, TypeVar

from qmi.core.transport import QMI_Transport
from qmi.core.exceptions import QMI_InstrumentException

# APT format specifiers
apt_word = c_uint16
apt_short = c_int16
apt_dword = c_uint32
apt_long = c_int32
apt_char = c_char
# this format specifier is not defined in the APT protocol manual but is helpful for packets that are divided into
# single bytes
apt_byte = c_uint8


class AptStatusBits(Enum):
    """Status bits for a status update message."""

    P_MOT_SB_CWHARDLIMIT = 0x00000001  # clockwise hardware limit switch
    P_MOT_SB_CCWHARDLIMIT = 0x00000002  # counter clockwise hardware limit switch
    P_MOT_SB_CWSOFTLIMIT = 0x00000004  # clockwise software limit switch
    P_MOT_SB_CCWSOFTLIMIT = 0x00000008  # counter clockwise software limit switch
    P_MOT_SB_INMOTIONCW = 0x00000010  # in motion, clockwise direction
    P_MOT_SB_INMOTIONCCW = 0x00000020  # in motion, counter clockwise direction
    P_MOT_SB_JOGGINGCW = 0x00000040  # jogging in clockwise direction
    P_MOT_SB_JOGGINGCCW = 0x00000080  # jogging in counter clockwise direction
    P_MOT_SB_CONNECTED = 0x00000100  # motor recognised by controller
    P_MOT_SB_HOMING = 0x00000200  # motor is homing
    P_MOT_SB_HOMED = 0x00000400  # motor is homed
    P_MOT_SB_INITIALISING = 0x00000800  # motor performing phase initialisation
    P_MOT_SB_TRACKING = 0x00001000  # actual position is within the tracking window
    P_MOT_SB_SETTLED = 0x00002000  # motor not moving and at target position
    P_MOT_SB_POSITIONERERROR = 0x00004000  # actual position outside margin specified around trajectory position
    P_MOT_SB_INSTRERROR = 0x00008000  # unable to execute command
    P_MOT_SB_INTERLOCK = 0x00010000  # used in controllers where a seperate signal is used to enable the motor
    P_MOT_SB_OVERTEMP = 0x00020000  # motor or motor power driver electronics reached maximum temperature
    P_MOT_SB_BUSVOLTFAULT = 0x00040000  # low supply voltage
    P_MOT_SB_COMMUTATIONERROR = 0x00080000  # problem with motor commutation. Can only be recovered with power cycle
    P_MOT_SB_DIGIP1 = 0x00100000  # state of digital input 1
    P_MOT_SB_DIGIP2 = 0x00200000  # state of digital input 2
    P_MOT_SB_DIGIP4 = 0x00400000  # state of digital input 3
    P_MOT_SB_DIGIP8 = 0x00800000  # state of digital input 4
    P_MOT_SB_OVERLOAD = 0x01000000  # some form of motor overload
    P_MOT_SB_ENCODERFAULT = 0x02000000  # encoder fault
    P_MOT_SB_OVERCURRENT = 0x04000000  # motor exceeded continuous current limit
    P_MOT_SB_BUSCURRENTFAULT = 0x08000000  # excessive current being drawn from motor power supply
    P_MOT_SB_POWEROK = 0x10000000  # controller power supplies operating normally
    P_MOT_SB_ACTIVE = 0x20000000  # controller executing motion commend
    P_MOT_SB_ERROR = 0x40000000  # indicates an error condition
    P_MOT_SB_ENABLED = 0x80000000  # motor output enabled, with controller maintaining position


class AptMessageId(Enum):
    """Message IDs for devices using the APT protocol."""

    HW_REQ_INFO = 0x0005
    HW_GET_INFO = 0x0006
    MOD_IDENTIFY = 0x0223
    MOD_SET_CHANENABLESTATE = 0x0210
    MOD_REQ_CHANENABLESTATE = 0x0211
    MOD_GET_CHANENABLESTATE = 0x0212
    HW_START_UPDATEMSGS = 0x0011
    HW_STOP_UPDATEMSGS = 0x0012
    MOT_MOVE_HOME = 0x0443
    MOT_MOVE_HOMED = 0x0444
    MOT_MOVE_ABSOLUTE = 0x0453
    MOT_MOVE_COMPLETED = 0x0464
    MOT_SET_EEPROMPARAMS = 0x04B9
    MOT_REQ_USTATUSUPDATE = 0x0490
    MOT_GET_USTATUSUPDATE = 0x0491
    MOT_MOVE_JOG = 0x046A
    POL_SET_PARAMS = 0x0530
    POL_REQ_PARAMS = 0x0531
    POL_GET_PARAMS = 0x0532


class AptChannelState(Enum):
    """Channel state"""

    ENABLE = 0x01
    DISABLE = 0x02


class AptChannelJogDirection(Enum):
    """Jog direction"""

    FORWARD = 0x01
    BACKWARD = 0x02


class AptMessageHeaderWithParams(LittleEndianStructure):
    """
    This is the version of the APT message header when no data packet follows a header.
    """

    _pack_ = True
    _fields_: ClassVar[list[tuple[str, type]]] = [
        ("message_id", c_uint16),
        ("param1", c_uint8),
        ("param2", c_uint8),
        ("dest", c_uint8),
        ("source", c_uint8),
    ]


class AptMessageHeaderForData(LittleEndianStructure):
    """
    This is the version of the APT message header when a data packet follows a header.
    """

    _pack_ = True
    _fields_: ClassVar[list[tuple[str, type]]] = [
        ("message_id", c_uint16),
        ("data_length", c_uint16),
        ("dest", c_uint8),
        ("source", c_uint8),
    ]


class AptMessage(LittleEndianStructure):
    """
    Base class for an APT message.
    """

    MESSAGE_ID: int
    HEADER_ONLY: bool = False
    _pack_ = True


T = TypeVar("T", bound=AptMessage)


class AptProtocol:
    """
    Implement the Thorlabs APT protocol primitives.
    """

    HEADER_SIZE_BYTES = 6

    def __init__(
        self,
        transport: QMI_Transport,
        apt_device_address: int = 0x50,
        host_address: int = 0x01,
        default_timeout: float | None = None,
    ):
        """Initialize the Thorlabs APT protocol handler.

        Parameters:
            transport:          Instance of `QMI_Transport` to use for sending APT commands to the instrument.
            apt_device_address: The address of the APT device. By default it is 0x50 which is a generic USB hardware
                                unit.
            host_address:       The address of the host that sends and receives messages. By default it is 0x01 which
                                is a host controller such as a PC.
            default_timeout:    Optional default response timeout in seconds.
                                The default is to wait indefinitely until a response is received.
        """
        self._transport = transport
        self._timeout = default_timeout
        self._apt_device_address = apt_device_address
        self._host_address = host_address

    def write_param_command(
        self,
        message_id: int,
        param1: int = 0x00,
        param2: int = 0x00,
    ) -> None:
        """
        Send an APT protocol command that is a header (i.e. 6 bytes) with params.

        Parameters:
            message_id: ID of message to send.
            param1:     Parameter 1 to be sent with default value of 0x00.
            param2:     Parameter 2 to be sent with default value of 0x00.
        """
        # Make the command.
        msg = AptMessageHeaderWithParams(
            message_id,
            param1,
            param2,
            self._apt_device_address,
            self._host_address,
        )
        # Send command.
        self._transport.write(bytearray(msg))

    def write_data_command(self, message_id: int, data: T) -> None:
        """
        Send and APT protocol command with data.
        """
        # Get size of data packet.
        data_length = sizeof(data)

        # Make the header.
        msg = AptMessageHeaderForData(message_id, data_length, self._apt_device_address | 0x80, self._host_address)

        # Send the header and data packet.
        self._transport.write(bytearray(msg) + bytearray(data))

    def ask(self, data_type: Type[T], timeout: float | None = None) -> T:
        """
        Ask for a response.

        Parameters:
            data_type:  Data type of the data packet that follows the header.
            timeout:    Optional response timeout in seconds.

        Returns:
            The requested data typed as the provided data type.
        """

        if timeout is None:
            timeout = self._timeout

        # Read the header first.
        header_bytes = self._transport.read(nbytes=self.HEADER_SIZE_BYTES, timeout=timeout)
        if data_type.HEADER_ONLY:
            return data_type.from_buffer_copy(header_bytes)
        header = AptMessageHeaderForData.from_buffer_copy(header_bytes)
        data_length = header.data_length

        # Read the data packet that follows the header.
        data_bytes = self._transport.read(nbytes=data_length, timeout=timeout)

        # Check that the received message ID is the ID that is expected.
        if data_type.MESSAGE_ID != header.message_id:
            raise QMI_InstrumentException(
                f"Expected message with ID {data_type.MESSAGE_ID}, but received {header.message_id}"
            )

        return data_type.from_buffer_copy(data_bytes)
